from typing import TYPE_CHECKING

import boto3
import pytest
from dagster import asset, materialize, open_pipes_session
from dagster._core.execution.context.compute import AssetExecutionContext
from dagster._core.instance_for_test import instance_for_test
from dagster._core.pipes.utils import PipesEnvContextInjector

from dagster_aws.pipes import PipesCloudWatchMessageReader, PipesEMRClient, PipesS3MessageReader
from dagster_aws_tests.pipes_tests.utils import _MOTO_SERVER_URL, _S3_TEST_BUCKET

if TYPE_CHECKING:
    from mypy_boto3_emr import EMRClient


@pytest.fixture
def emr_client(moto_server) -> "EMRClient":
    return boto3.client("emr", region_name="us-east-1", endpoint_url=_MOTO_SERVER_URL)


@pytest.fixture
def pipes_emr_client(emr_client, s3_client) -> PipesEMRClient:
    return PipesEMRClient(
        client=emr_client,
        message_reader=PipesS3MessageReader(bucket=_S3_TEST_BUCKET, client=s3_client),
    )


def test_emr_enrich_params(pipes_emr_client: "PipesEMRClient"):
    @asset
    def my_asset(context: AssetExecutionContext, pipes_emr_client: PipesEMRClient):
        message_reader = PipesCloudWatchMessageReader()
        context_injector = PipesEnvContextInjector()

        with open_pipes_session(
            context=context,
            message_reader=message_reader,
            context_injector=context_injector,
        ) as session:
            params = pipes_emr_client._enrich_params(  # noqa: SLF001
                session=session,
                params={
                    "Name": "test",
                    "ReleaseLabel": "emr-7.2.0-latest",
                    "LogUri": "s3://my-bucket/logs",
                    "Instances": {
                        "MasterInstanceType": "m5.xlarge",
                        "SlaveInstanceType": "m5.xlarge",
                        "InstanceCount": 2,
                    },
                    "Steps": [
                        {
                            "Name": "test",
                            "ActionOnFailure": "CONTINUE",
                            "HadoopJarStep": {
                                "Jar": "command-runner.jar",
                                "Args": ["spark-submit", "--version"],
                            },
                        }
                    ],
                    "Configurations": [],
                },
            )

            run_id_tag_injected = False

            for tag in params.get("Tags", []):
                if tag.get("Key") == "dagster/run-id":
                    run_id_tag_injected = True
                    assert tag.get("Value") == context.run_id

            assert run_id_tag_injected

            return session.get_results()

    with instance_for_test() as instance:
        materialize(
            [my_asset],
            resources={"pipes_emr_client": pipes_emr_client},
            instance=instance,
        )



from dagster import op
from dagster._utils.test import wrap_op_in_graph_and_execute

from dagster_aws.ecr import FakeECRPublicResource, fake_ecr_public_resource


def test_ecr_public_get_login_password() -> None:
    @op(required_resource_keys={"ecr_public"})
    def ecr_public_op(context):
        return context.resources.ecr_public.get_login_password()

    result = wrap_op_in_graph_and_execute(
        ecr_public_op,
        resources={"ecr_public": fake_ecr_public_resource},
    )

    assert result.output_value() == "token"


def test_ecr_public_get_login_password_pythonic() -> None:
    @op
    def ecr_public_op(ecr_public: FakeECRPublicResource):
        return ecr_public.get_client().get_login_password()

    result = wrap_op_in_graph_and_execute(
        ecr_public_op,
        resources={"ecr_public": FakeECRPublicResource()},
    )

    assert result.output_value() == "token"



import pytest

from dagster_aws.ecs.utils import RetryableEcsException, run_ecs_task


def test_run_task_failure(ecs, instance, workspace, run):
    def run_task(self=ecs, **kwargs):
        self.stubber.activate()
        self.stubber.add_response(
            method="run_task",
            service_response={
                "tasks": [],
                "failures": [
                    {"arn": "failing-arn-1", "reason": "boom", "detail": "detailed boom 1"},
                    {"arn": "missing-detail", "reason": "too succinct"},
                    {"reason": "ran out of arns"},
                ],
            },
            expected_params={**kwargs},
        )
        response = self.client.run_task(**kwargs)
        self.stubber.deactivate()
        return response

    instance.run_launcher.ecs.run_task = run_task

    with pytest.raises(Exception) as ex:
        instance.launch_run(run.run_id, workspace)

    assert ex.match(
        "Task failing-arn-1 failed. Failure reason: boom Failure details: detailed boom 1\n"
    )
    assert ex.match("\nTask missing-detail failed. Failure reason: too succinct\n")
    assert ex.match("Task failed. Failure reason: ran out of arns")


def test_run_task_retryrable_failure(ecs, instance, workspace, run, other_run, monkeypatch):
    original = ecs.run_task

    out_of_capacity_response = {
        "tasks": [],
        "failures": [
            {
                "arn": "missing-capacity",
                "reason": "Capacity is unavailable at this time. Please try again later or in a different availability zone",
                "detail": "boom",
            },
        ],
    }

    retryable_failures = iter([out_of_capacity_response])

    def run_task(*args, **kwargs):
        try:
            return next(retryable_failures)
        except StopIteration:
            return original(*args, **kwargs)

    instance.run_launcher.ecs.run_task = run_task

    instance.launch_run(run.run_id, workspace)

    # reset our mock and test again with 0 retries
    retryable_failures = iter([out_of_capacity_response])

    monkeypatch.setenv("RUN_TASK_RETRIES", "0")

    with pytest.raises(Exception) as ex:
        instance.launch_run(other_run.run_id, workspace)

    assert ex.match("Capacity is unavailable")


def test_run_task_throttle_failure(ecs, instance, workspace, run, other_run, monkeypatch):
    original = ecs.run_task

    throttle_error = ecs.exceptions.InvalidParameterException(
        error_response={
            "Error": {
                "Code": "InvalidParameterException",
                "Message": (
                    "Error retrieving subnet information for [subnet-abc123]:"
                    " Rate exceeded (ErrorCode: Throttling)"
                ),
            }
        },
        operation_name="RunTask",
    )

    throttle_failures = iter([throttle_error])

    def run_task(*args, **kwargs):
        try:
            raise next(throttle_failures)
        except StopIteration:
            # Falls through to the real run_task
            return original(*args, **kwargs)

    instance.run_launcher.ecs.run_task = run_task

    # First call raises throttle error, retry succeeds
    instance.launch_run(run.run_id, workspace)

    # Reset our mock and test again with 0 retries
    throttle_failures = iter([throttle_error])

    monkeypatch.setenv("RUN_TASK_RETRIES", "0")

    with pytest.raises(RetryableEcsException, match="Throttling"):
        instance.launch_run(other_run.run_id, workspace)


def test_run_task_non_throttle_invalid_parameter(ecs):
    """Non-throttling InvalidParameterException should NOT be retried."""
    non_throttle_error = ecs.exceptions.InvalidParameterException(
        error_response={
            "Error": {
                "Code": "InvalidParameterException",
                "Message": "Invalid parameter value",
            }
        },
        operation_name="RunTask",
    )

    class FakeEcs:
        exceptions = ecs.exceptions

        def run_task(self, **_kwargs):
            raise non_throttle_error

    with pytest.raises(ecs.exceptions.InvalidParameterException, match="Invalid parameter value"):
        run_ecs_task(FakeEcs(), {})



import copy
import json
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Any

import pytest
from botocore.exceptions import ClientError
from dagster._check import CheckError
from dagster._core.code_pointer import FileCodePointer
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.launcher import LaunchRunContext
from dagster._core.launcher.base import WorkerStatus
from dagster._core.origin import JobPythonOrigin, RepositoryPythonOrigin
from dagster._core.storage.dagster_run import DagsterRunStatus
from dagster._core.storage.tags import HIDDEN_TAG_PREFIX, RUN_WORKER_ID_TAG

import dagster_aws
from dagster_aws.ecs import EcsEventualConsistencyTimeout
from dagster_aws.ecs.launcher import (
    DEFAULT_LINUX_RESOURCES,
    DEFAULT_WINDOWS_RESOURCES,
    RUNNING_STATUSES,
    STOPPED_STATUSES,
)
from dagster_aws.ecs.tasks import DagsterEcsTaskDefinitionConfig
from dagster_aws.ecs.utils import get_task_definition_family


@pytest.mark.parametrize("task_long_arn_format", ["enabled", "disabled"])
def test_default_launcher(
    ecs,
    instance,
    workspace,
    run,
    subnet,
    image,
    environment,
    task_long_arn_format,
):
    ecs.put_account_setting(name="taskLongArnFormat", value=task_long_arn_format)
    assert "ecs/task_arn" not in run.tags
    assert "ecs/cluster" not in run.tags

    initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    initial_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    # A new task definition is created
    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    assert len(task_definitions) == len(initial_task_definitions) + 1
    task_definition_arn = next(iter(set(task_definitions).difference(initial_task_definitions)))
    task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)
    task_definition = task_definition["taskDefinition"]

    assert instance.run_launcher.check_run_worker_health(run).status == WorkerStatus.RUNNING

    # It has a new family, name, and image
    assert task_definition["family"] == get_task_definition_family("run", run.remote_job_origin)
    assert len(task_definition["containerDefinitions"]) == 1
    container_definition = task_definition["containerDefinitions"][0]
    assert container_definition["name"] == "run"
    assert container_definition["image"] == image
    assert not container_definition.get("entryPoint")
    assert not container_definition.get("dependsOn")
    # But other stuff is inherited from the parent task definition
    assert all(item in container_definition["environment"] for item in environment)
    assert {"name": "DAGSTER_RUN_JOB_NAME", "value": "job"} in container_definition["environment"]

    # A new task is launched
    tasks = ecs.list_tasks()["taskArns"]

    assert len(tasks) == len(initial_tasks) + 1
    task_arn = next(iter(set(tasks).difference(initial_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]
    assert subnet.id in str(task)
    assert task["taskDefinitionArn"] == task_definition["taskDefinitionArn"]
    assert task["launchType"] == "FARGATE"

    # The run is tagged with info about the ECS task
    assert instance.get_run_by_id(run.run_id).tags["ecs/task_arn"] == task_arn
    cluster_arn = ecs._cluster_arn("default")  # noqa: SLF001
    assert instance.get_run_by_id(run.run_id).tags["ecs/cluster"] == cluster_arn

    # If we're using the new long ARN format,
    # the ECS task is tagged with info about the Dagster run
    if task_long_arn_format == "enabled":
        assert (
            ecs.list_tags_for_resource(resourceArn=task_arn)["tags"][0]["key"] == "dagster/run_id"
        )
        assert ecs.list_tags_for_resource(resourceArn=task_arn)["tags"][0]["value"] == run.run_id

        assert (
            ecs.list_tags_for_resource(resourceArn=task_arn)["tags"][1]["key"] == "dagster/job_name"
        )
        assert ecs.list_tags_for_resource(resourceArn=task_arn)["tags"][1]["value"] == run.job_name

    # We set job-specific overides
    overrides = task["overrides"]["containerOverrides"]
    assert len(overrides) == 1
    override = overrides[0]
    assert override["name"] == "run"
    assert "execute_run" in override["command"]
    assert run.run_id in str(override["command"])

    # And we log
    events = instance.event_log_storage.get_logs_for_run(run.run_id)
    latest_event = events[-1]
    assert latest_event.message == "[EcsRunLauncher] Launching run in ECS task"
    metadata = latest_event.dagster_event.engine_event_data.metadata
    assert metadata["ECS Task ARN"] == MetadataValue.text(task_arn)
    assert metadata["ECS Cluster"] == MetadataValue.text(cluster_arn)
    assert metadata["Run ID"] == MetadataValue.text(run.run_id)

    # check status and stop task
    assert instance.run_launcher.check_run_worker_health(run).status == WorkerStatus.RUNNING
    ecs.stop_task(task=task_arn)


def test_launcher_fargate_spot(
    ecs,
    instance_fargate_spot,
    workspace,
    remote_job,
    job,
    subnet,
    image,
    environment,
):
    instance = instance_fargate_spot
    run = instance.create_run_for_job(
        job,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
    )
    initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    initial_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    # A new task definition is still created
    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    assert len(task_definitions) == len(initial_task_definitions) + 1
    task_definition_arn = next(iter(set(task_definitions).difference(initial_task_definitions)))
    task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)
    task_definition = task_definition["taskDefinition"]

    # A new task is launched
    tasks = ecs.list_tasks()["taskArns"]

    assert len(tasks) == len(initial_tasks) + 1
    task_arn = next(iter(set(tasks).difference(initial_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert task["capacityProviderName"] == "FARGATE_SPOT"

    # Override capacity provider strategy with tags
    run = instance.create_run_for_job(
        job,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
    )
    instance.add_run_tags(
        run.run_id,
        {
            "ecs/run_task_kwargs": json.dumps(
                {
                    "capacityProviderStrategy": [
                        {
                            "capacityProvider": "CUSTOM",
                        },
                    ],
                }
            )
        },
    )
    instance.launch_run(run.run_id, workspace)

    # A new task definition is still created
    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    assert len(task_definitions) == len(initial_task_definitions) + 1
    task_definition_arn = next(iter(set(task_definitions).difference(initial_task_definitions)))
    task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)
    task_definition = task_definition["taskDefinition"]

    # A new task is launched
    second_tasks = ecs.list_tasks()["taskArns"]

    assert len(second_tasks) == len(tasks) + 1
    task_arn = next(iter(set(second_tasks).difference(tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert task["capacityProviderName"] == "CUSTOM"


def test_launcher_dont_use_current_task(
    ecs,
    instance_dont_use_current_task,
    workspace,
    remote_job,
    job,
    subnet,
    image,
    environment,
):
    instance = instance_dont_use_current_task
    run = instance.create_run_for_job(
        job,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
    )

    cluster = instance.run_launcher.run_task_kwargs["cluster"]
    assert cluster == "my_cluster"

    assert "ecs/task_arn" not in run.tags
    assert "ecs/cluster" not in run.tags

    initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    initial_tasks = ecs.list_tasks(cluster=cluster)["taskArns"]

    instance.launch_run(run.run_id, workspace)

    # A new task definition is still created
    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    assert len(task_definitions) == len(initial_task_definitions) + 1
    task_definition_arn = next(iter(set(task_definitions).difference(initial_task_definitions)))
    task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)
    task_definition = task_definition["taskDefinition"]

    assert instance.run_launcher.check_run_worker_health(run).status == WorkerStatus.RUNNING

    # It has a new family, name, and image
    assert task_definition["family"] == get_task_definition_family("run", run.remote_job_origin)
    assert len(task_definition["containerDefinitions"]) == 1
    container_definition = task_definition["containerDefinitions"][0]
    assert container_definition["name"] == "run"
    assert container_definition["image"] == image
    assert not container_definition.get("entryPoint")
    assert not container_definition.get("dependsOn")

    # It does not take in the environment configured on the calling task definition
    assert not any(item in container_definition["environment"] for item in environment)
    assert {"name": "DAGSTER_RUN_JOB_NAME", "value": "job"} in container_definition["environment"]

    # A new task is launched
    tasks = ecs.list_tasks(cluster=cluster)["taskArns"]

    assert len(tasks) == len(initial_tasks) + 1
    task_arn = next(iter(set(tasks).difference(initial_tasks)))
    task = ecs.describe_tasks(cluster=cluster, tasks=[task_arn])["tasks"][0]
    assert subnet.id in str(task)
    assert task["taskDefinitionArn"] == task_definition["taskDefinitionArn"]

    # The run is tagged with info about the ECS task
    assert instance.get_run_by_id(run.run_id).tags["ecs/task_arn"] == task_arn
    cluster_arn = ecs._cluster_arn(cluster)  # noqa: SLF001
    assert instance.get_run_by_id(run.run_id).tags["ecs/cluster"] == cluster_arn

    assert ecs.list_tags_for_resource(resourceArn=task_arn)["tags"][0]["key"] == "dagster/run_id"
    assert ecs.list_tags_for_resource(resourceArn=task_arn)["tags"][0]["value"] == run.run_id

    # We set job-specific overides
    overrides = task["overrides"]["containerOverrides"]
    assert len(overrides) == 1
    override = overrides[0]
    assert override["name"] == "run"
    assert "execute_run" in override["command"]
    assert run.run_id in str(override["command"])

    # And we log
    events = instance.event_log_storage.get_logs_for_run(run.run_id)
    latest_event = events[-1]
    assert latest_event.message == "[EcsRunLauncher] Launching run in ECS task"
    metadata = latest_event.dagster_event.engine_event_data.metadata
    assert metadata["ECS Task ARN"] == MetadataValue.text(task_arn)
    assert metadata["ECS Cluster"] == MetadataValue.text(cluster_arn)
    assert metadata["Run ID"] == MetadataValue.text(run.run_id)

    # check status and stop task
    assert instance.run_launcher.check_run_worker_health(run).status == WorkerStatus.RUNNING
    ecs.stop_task(cluster=cluster, task=task_arn)


def test_task_definition_registration(
    ecs, instance, workspace, run, other_workspace, other_run, secrets_manager, monkeypatch
):
    initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    _initial_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    # A new task definition is created
    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    assert len(task_definitions) == len(initial_task_definitions) + 1

    # Launching another run reuses an existing task definition
    instance.launch_run(run.run_id, workspace)
    assert task_definitions == ecs.list_task_definitions()["taskDefinitionArns"]

    # Register a new task definition if the image changes
    instance.launch_run(other_run.run_id, other_workspace)
    assert len(ecs.list_task_definitions()["taskDefinitionArns"]) == len(task_definitions) + 1

    # Relaunching another run with the new image reuses an existing task definition
    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    instance.launch_run(other_run.run_id, other_workspace)
    assert task_definitions == ecs.list_task_definitions()["taskDefinitionArns"]

    # Register a new task definition if secrets change
    secrets_manager.create_secret(
        Name="hello",
        SecretString="hello",
        Tags=[{"Key": "dagster", "Value": "true"}],
    )

    instance.launch_run(other_run.run_id, other_workspace)
    assert len(ecs.list_task_definitions()["taskDefinitionArns"]) == len(task_definitions) + 1

    # Relaunching another run with the same secrets reuses an existing task definition
    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    instance.launch_run(other_run.run_id, other_workspace)
    assert task_definitions == ecs.list_task_definitions()["taskDefinitionArns"]

    # Register a new task definition if _reuse_task_definition returns False
    # for any other reason
    monkeypatch.setattr(instance.run_launcher, "_reuse_task_definition", lambda *_: False)

    instance.launch_run(other_run.run_id, other_workspace)
    assert len(ecs.list_task_definitions()["taskDefinitionArns"]) == len(task_definitions) + 1


@pytest.mark.skip(
    "This remains occasionally flaky on older versions of Python. See"
    " https://github.com/dagster-io/dagster/pull/11290 "
    "https://linear.app/elementl/issue/CLOUD-2093/re-enable-flaky-ecs-task-registration-race-condition-tests"
)
def test_task_definition_registration_race_condition(ecs, instance, workspace, run):
    initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    initial_tasks = ecs.list_tasks()["taskArns"]

    with ThreadPoolExecutor(max_workers=10) as executor:
        for i in range(10):
            executor.submit(instance.launch_run, run.run_id, workspace)

    task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
    assert len(task_definitions) == len(initial_task_definitions) + 1

    tasks = ecs.list_tasks()["taskArns"]
    assert len(tasks) == len(initial_tasks) + 10


def test_reuse_task_definition(instance, ecs):
    image = "image"
    secrets = []
    environment = [
        {
            "name": "MY_ENV_VAR",
            "value": "MY_VALUE",
        },
        {
            "name": "MY_OTHER_ENV_VAR",
            "value": "MY_OTHER_VALUE",
        },
    ]

    container_name = instance.run_launcher.container_name

    original_task_definition = {
        "family": "hello",
        "containerDefinitions": [
            {
                "image": image,
                "name": container_name,
                "secrets": secrets,
                "environment": environment,
                "command": ["echo", "HELLO"],
                "healthCheck": {
                    "command": ["HELLO"],
                },
            },
            {
                "image": "other_image",
                "name": "the_sidecar",
            },
        ],
        "cpu": "256",
        "memory": "512",
        "taskRoleArn": "arn:aws:iam::12345:role/dagstertest-TaskRole-kQcbapXu8MZg",
        "executionRoleArn": "arn:aws:iam::12345:role/dagstertest-ExecutionRole-kQcbapXu8MZg",
    }

    task_definition_config = DagsterEcsTaskDefinitionConfig.from_task_definition_dict(
        original_task_definition, container_name
    )

    # New task definition not re-used since it is new
    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        task_definition_config, container_name
    )
    # Once it's registered, it is re-used

    ecs.register_task_definition(**original_task_definition)

    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        task_definition_config, container_name
    )

    # Reordering environment is still reused
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["environment"] = list(
        reversed(task_definition["containerDefinitions"][0]["environment"])
    )
    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed image fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["image"] = "new-image"

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed container name fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["name"] = "new-container"
    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, "new-container"),
        container_name,
    )

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(
            original_task_definition, container_name
        ),
        "new-container",
    )

    # Changed secrets fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["secrets"].append(
        {"name": "new-secret", "valueFrom": "fake-arn"}
    )
    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed environment fails

    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["environment"].append(
        {"name": "MY_ENV_VAR", "value": "MY_ENV_VALUE"}
    )
    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(
            task_definition,
            container_name,
        ),
        container_name,
    )

    # Changed execution role fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["executionRoleArn"] = (
        "arn:aws:iam::12345:role/dagstertest-OtherExecutionRole-kQcbapXu8MZg"
    )
    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed task role fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["taskRoleArn"] = (
        "arn:aws:iam::12345:role/dagstertest-OtherTaskRole-kQcbapXu8MZg"
    )
    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed runtime platform does not fail
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["runtimePlatform"] = {"operatingSystemFamily": "WINDOWS_SERVER_2019_FULL"}
    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed command fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["command"] = ["echo", "GOODBYE"]
    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed linuxParameters fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["linuxParameters"] = {
        "capabilities": {"add": ["SYS_PTRACE"]},
        "initProcessEnabled": True,
    }
    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Changed healthCheck does not fail
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["healthCheck"] = {
        "command": ["CMD-SHELL", "curl -f http://localhost/ || exit 1"],
        "interval": 30,
        "timeout": 5,
        "retries": 3,
        "startPeriod": 0,
    }
    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Any other diff passes
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["somethingElse"] = "boom"
    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Different sidecar image fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][1]["image"] = "new_sidecar_image"

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Different sidecar name fails
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][1]["name"] = "new_sidecar_name"

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Different sidecar environments fail
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][1]["environment"] = environment

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Different sidecar secrets fail
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][1]["secrets"] = [
        {"name": "a_secret", "valueFrom": "an_arn"}
    ]

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Other changes to sidecars do not fail
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][1]["cpu"] = "256"
    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Matches if the only difference between the task role arns is that they remove the arn prefix
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["taskRoleArn"] = "dagstertest-TaskRole-kQcbapXu8MZg"

    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Matches if the only difference between the execution role arns is that they remove the arn prefix
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["executionRoleArn"] = "dagstertest-ExecutionRole-kQcbapXu8MZg"

    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(task_definition, container_name),
        container_name,
    )

    # Fails if the existing task definition has a different container name
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["containerDefinitions"][0]["name"] = "foobar"
    ecs.register_task_definition(**task_definition)

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(
            original_task_definition, container_name
        ),
        container_name,
    )

    # Fails if the existing task definition has a different ephemeral storage
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["ephemeralStorage"] = {"sizeInGiB": 120}
    ecs.register_task_definition(**task_definition)

    assert not instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(
            original_task_definition, container_name
        ),
        container_name,
    )

    # Matches if the existing task definition uses the default ephemeralStorage
    # https://docs.aws.amazon.com/AmazonECS/latest/developerguide/fargate-task-storage.html
    task_definition = copy.deepcopy(original_task_definition)
    task_definition["ephemeralStorage"] = {"sizeInGiB": 20}
    ecs.register_task_definition(**task_definition)

    assert instance.run_launcher._reuse_task_definition(  # noqa: SLF001
        DagsterEcsTaskDefinitionConfig.from_task_definition_dict(
            original_task_definition, container_name
        ),
        container_name,
    )


def test_task_definition_prefix(ecs, instance_cm, run, workspace, job, remote_job):
    with instance_cm(
        {
            "task_definition_prefix": "foo",
        }
    ) as instance:
        initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]

        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )
        instance.launch_run(run.run_id, workspace)

        task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
        assert len(task_definitions) == len(initial_task_definitions) + 1
        task_definition_arn = next(iter(set(task_definitions).difference(initial_task_definitions)))
        task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)
        task_definition = task_definition["taskDefinition"]

        assert task_definition["family"] == get_task_definition_family("foo", run.remote_job_origin)


def test_default_task_definition_resources(ecs, instance_cm, run, workspace, job, remote_job):
    task_role_arn = "fake-task-role"
    execution_role_arn = "fake-execution-role"
    with instance_cm(
        {
            "task_definition": {
                "task_role_arn": task_role_arn,
                "execution_role_arn": execution_role_arn,
            },
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )
        initial_tasks = ecs.list_tasks()["taskArns"]

        instance.launch_run(run.run_id, workspace)

        tasks = ecs.list_tasks()["taskArns"]
        task_arn = next(iter(set(tasks).difference(initial_tasks)))

        task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

        task_definition_arn = task["taskDefinitionArn"]

        task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)[
            "taskDefinition"
        ]

        # Since this is not a windows task def, the default cpu/memory are 256/512
        assert task_definition["cpu"] == DEFAULT_LINUX_RESOURCES["cpu"]
        assert task_definition["memory"] == DEFAULT_LINUX_RESOURCES["memory"]

    # But a windows task def has different defaults that are valid for window
    with instance_cm(
        {
            "task_definition": {
                "task_role_arn": task_role_arn,
                "execution_role_arn": execution_role_arn,
                "runtime_platform": {"operatingSystemFamily": "WINDOWS_SERVER_2019_FULL"},
            },
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )
        initial_tasks = ecs.list_tasks()["taskArns"]

        instance.launch_run(run.run_id, workspace)

        tasks = ecs.list_tasks()["taskArns"]
        task_arn = next(iter(set(tasks).difference(initial_tasks)))

        task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

        task_definition_arn = task["taskDefinitionArn"]

        task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)[
            "taskDefinition"
        ]

        # Default cpu/memory is higher on windows
        assert task_definition["cpu"] == DEFAULT_WINDOWS_RESOURCES["cpu"]
        assert task_definition["memory"] == DEFAULT_WINDOWS_RESOURCES["memory"]

    # Setting resources on the run launcher ignores the defaults
    with instance_cm(
        {
            "task_definition": {
                "task_role_arn": task_role_arn,
                "execution_role_arn": execution_role_arn,
            },
            "run_resources": {"cpu": "2048", "memory": "4096", "ephemeral_storage": 36},
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )
        initial_tasks = ecs.list_tasks()["taskArns"]

        instance.launch_run(run.run_id, workspace)

        tasks = ecs.list_tasks()["taskArns"]
        task_arn = next(iter(set(tasks).difference(initial_tasks)))

        task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

        task_definition_arn = task["taskDefinitionArn"]

        task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)[
            "taskDefinition"
        ]

        assert task_definition["cpu"] == "2048"
        assert task_definition["memory"] == "4096"
        assert task_definition["ephemeralStorage"]["sizeInGiB"] == 36


def test_launching_with_task_definition_dict(ecs, instance_cm, run, workspace, job, remote_job):
    container_name = "dagster"

    task_role_arn = "fake-task-role"
    execution_role_arn = "fake-execution-role"
    sidecar = {
        "name": "DatadogAgent",
        "image": "public.ecr.aws/datadog/agent:latest",
        "environment": [
            {"name": "ECS_FARGATE", "value": "true"},
        ],
    }
    log_group = "my-log-group"

    mount_points = [
        {
            "sourceVolume": "myEfsVolume",
            "containerPath": "/mount/efs",
            "readOnly": True,
        }
    ]
    volumes = [
        {
            "name": "myEfsVolume",
            "efsVolumeConfiguration": {
                "fileSystemId": "fs-1234",
                "rootDirectory": "/path/to/my/data",
            },
        }
    ]

    repository_credentials = "fake-secret-arn"

    linux_parameters = {
        "capabilities": {"add": ["SYS_PTRACE"]},
        "initProcessEnabled": True,
    }

    # You can provide a family or a task definition ARN
    with instance_cm(
        {
            "task_definition": {
                "log_group": log_group,
                "task_role_arn": task_role_arn,
                "execution_role_arn": execution_role_arn,
                "sidecar_containers": [sidecar],
                "requires_compatibilities": ["FARGATE"],
                "runtime_platform": {"operatingSystemFamily": "WINDOWS_SERVER_2019_FULL"},
                "mount_points": mount_points,
                "volumes": volumes,
                "repository_credentials": repository_credentials,
                "linux_parameters": linux_parameters,
            },
            "container_name": container_name,
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )

        initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
        initial_tasks = ecs.list_tasks()["taskArns"]

        instance.launch_run(run.run_id, workspace)

        new_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]

        # A new task definition is created
        assert new_task_definitions != initial_task_definitions

        # A new task is launched
        tasks = ecs.list_tasks()["taskArns"]
        assert len(tasks) == len(initial_tasks) + 1
        task_arn = next(iter(set(tasks).difference(initial_tasks)))
        task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]
        task_definition_arn = task["taskDefinitionArn"]

        task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)[
            "taskDefinition"
        ]

        assert task_definition["volumes"] == volumes
        assert task_definition["taskRoleArn"] == task_role_arn
        assert task_definition["executionRoleArn"] == execution_role_arn
        assert task_definition["runtimePlatform"] == {
            "operatingSystemFamily": "WINDOWS_SERVER_2019_FULL"
        }

        container_definition = task_definition["containerDefinitions"][0]
        assert container_definition["mountPoints"] == mount_points

        assert (
            container_definition["repositoryCredentials"]["credentialsParameter"]
            == repository_credentials
        )

        assert container_definition["linuxParameters"] == linux_parameters

        assert [container["name"] for container in task_definition["containerDefinitions"]] == [
            container_name,
            sidecar["name"],
        ]

        # We set job-specific overides
        overrides = task["overrides"]["containerOverrides"]
        assert len(overrides) == 1
        override = overrides[0]
        assert override["name"] == container_name
        assert "execute_run" in override["command"]
        assert run.run_id in str(override["command"])

        second_run = run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )

        instance.launch_run(second_run.run_id, workspace)

        # A new task definition is not created
        assert ecs.list_task_definitions()["taskDefinitionArns"] == new_task_definitions


def test_launching_custom_task_definition(ecs, instance_cm, run, workspace, job, remote_job):
    container_name = "override_container"

    task_definition = ecs.register_task_definition(
        family="override",
        containerDefinitions=[{"name": container_name, "image": "hello_world:latest"}],
        networkMode="bridge",
        memory="512",
        cpu="256",
    )["taskDefinition"]
    task_definition_arn = task_definition["taskDefinitionArn"]
    family = task_definition["family"]

    # The task definition doesn't exist
    with pytest.raises(Exception), instance_cm({"task_definition": "does not exist"}) as instance:
        print(instance.run_launcher)  # noqa: T201

    # The task definition doesn't include the default container name
    with pytest.raises(CheckError), instance_cm({"task_definition": family}) as instance:
        print(instance.run_launcher)  # noqa: T201

    # The task definition doesn't include the container name
    with (
        pytest.raises(CheckError),
        instance_cm({"task_definition": family, "container_name": "does not exist"}) as instance,
    ):
        print(instance.run_launcher)  # noqa: T201

    # You can provide a family or a task definition ARN
    with instance_cm(
        {"task_definition": task_definition_arn, "container_name": container_name}
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )

        initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
        initial_tasks = ecs.list_tasks()["taskArns"]

        instance.launch_run(run.run_id, workspace)

        # A new task definition is not created
        assert ecs.list_task_definitions()["taskDefinitionArns"] == initial_task_definitions

        # A new task is launched
        tasks = ecs.list_tasks()["taskArns"]
        assert len(tasks) == len(initial_tasks) + 1
        task_arn = next(iter(set(tasks).difference(initial_tasks)))
        task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]
        assert task["taskDefinitionArn"] == task_definition["taskDefinitionArn"]

        # We set job-specific overides
        overrides = task["overrides"]["containerOverrides"]
        assert len(overrides) == 1
        override = overrides[0]
        assert override["name"] == container_name
        assert "execute_run" in override["command"]
        assert run.run_id in str(override["command"])


def test_eventual_consistency(ecs, instance, workspace, run, monkeypatch):
    initial_tasks = ecs.list_tasks()["taskArns"]

    retries = 0
    original_describe_tasks = instance.run_launcher.ecs.describe_tasks
    original_backoff_retries = dagster_aws.ecs.tasks.BACKOFF_RETRIES  # pyright: ignore[reportAttributeAccessIssue]

    def describe_tasks(*_args, **_kwargs):
        nonlocal retries
        nonlocal original_describe_tasks

        if retries > 1:
            return original_describe_tasks(*_args, **_kwargs)
        retries += 1
        return {"tasks": []}

    with pytest.raises(EcsEventualConsistencyTimeout):
        monkeypatch.setattr(instance.run_launcher.ecs, "describe_tasks", describe_tasks)
        monkeypatch.setattr(dagster_aws.ecs.tasks, "BACKOFF_RETRIES", 0)  # pyright: ignore[reportAttributeAccessIssue]
        instance.launch_run(run.run_id, workspace)

    # Reset the mock
    retries = 0
    monkeypatch.setattr(dagster_aws.ecs.tasks, "BACKOFF_RETRIES", original_backoff_retries)  # pyright: ignore[reportAttributeAccessIssue]
    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    assert len(tasks) == len(initial_tasks) + 1

    # backoff fails for reasons unrelated to eventual consistency

    def exploding_describe_tasks(*_args, **_kwargs):
        raise Exception("boom")

    with pytest.raises(Exception, match="boom"):
        monkeypatch.setattr(instance.run_launcher.ecs, "describe_tasks", exploding_describe_tasks)
        instance.launch_run(run.run_id, workspace)


@pytest.mark.parametrize("assign_public_ip", [True, False])
def test_public_ip_assignment(ecs, ec2, instance, workspace, run, assign_public_ip):
    initial_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(initial_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]
    attachment = task.get("attachments")[0]
    details = dict([(detail.get("name"), detail.get("value")) for detail in attachment["details"]])
    eni = ec2.NetworkInterface(details["networkInterfaceId"])
    attributes = eni.association_attribute or {}

    assert bool(attributes.get("PublicIp")) == assign_public_ip


def test_check_run_worker_health_adds_eni_tag(ecs, instance, workspace, run, monkeypatch):
    monkeypatch.setenv("DAGSTER_AWS_ENI_TAGGING_ENABLED", "true")

    initial_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(initial_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert not any(
        k.startswith(f"{HIDDEN_TAG_PREFIX}eni_id") for k in instance.get_run_by_id(run.run_id).tags
    )

    health = instance.run_launcher.check_run_worker_health(run)
    assert health.status == WorkerStatus.RUNNING

    attachment = task.get("attachments")[0]
    details = dict((detail.get("name"), detail.get("value")) for detail in attachment["details"])
    eni_id = details["networkInterfaceId"]

    run_tags = instance.get_run_by_id(run.run_id).tags
    assert run_tags.get(f"{HIDDEN_TAG_PREFIX}eni_id") == eni_id


def test_launcher_run_resources(
    ecs,
    instance_with_resources,
    workspace,
    remote_job,
    job,
):
    instance = instance_with_resources
    run = instance.create_run_for_job(
        job,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
    )

    existing_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(existing_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert task.get("overrides").get("memory") == "2048"
    assert task.get("overrides").get("cpu") == "1024"


def test_launch_cannot_use_system_tags(instance_cm, workspace, job, remote_job):
    with instance_cm(
        {
            "run_ecs_tags": [{"key": "dagster/run_id", "value": "NOPE"}],
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )
        with pytest.raises(Exception, match="Cannot override system ECS tag: dagster/run_id"):
            instance.launch_run(run.run_id, workspace)


@pytest.mark.parametrize(
    [
        "run_tags",
        "expected_ecs_tag_keys",
        "allow_list",
    ],
    [
        (
            {
                "dagster/partition_key": "abc",
                "dagster/git_commit_hash": "b54e4ddfbf2f4f661cdb312b6f3dd49de6139c94",
            },
            {"dagster/run_id", "dagster/job_name"},
            [],
        ),
        (
            {
                "dagster/partition_key": "abc",
                "dagster/git_commit_hash": "b54e4ddfbf2f4f661cdb312b6f3dd49de6139c94",
            },
            {"dagster/partition_key", "dagster/job_name", "dagster/run_id"},
            ["dagster/partition_key"],
        ),
    ],
)
def test_propagate_tags_include_all(
    run_tags: dict[str, str],
    expected_ecs_tag_keys: set[str],
    allow_list: list[str],
    instance_cm,
    workspace,
    job,
    remote_job,
    ecs,
):
    with instance_cm(
        {
            "propagate_tags": {
                "allow_list": allow_list,
            },
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )
        test_tags = {**run_tags, "ecs/memory": "30720", "ecs/cpu": "4096"}

        instance.add_run_tags(run.run_id, test_tags)
        instance.launch_run(run.run_id, workspace)

        run_tags.update({"dagster/run_id": run.run_id})
        initial_tasks = ecs.list_tasks()
        tasks = ecs.describe_tasks(tasks=initial_tasks["taskArns"])
        tags = tasks["tasks"][1]["tags"]
        expected_ecs_tag_keys.add("dagster/run_id")
        assert set([tag["key"] for tag in tags]) == expected_ecs_tag_keys


@pytest.mark.parametrize(
    ["propagate_tags"],
    [
        [{"allow_list": ["dagster/op_selection"]}],
        [{"allow_list": ["dagster/solid_selection", "dagster/git_commit_hash"]}],
    ],
)
def test_propagate_tags_args_validated(
    instance_cm,
    workspace,
    job,
    remote_job,
    propagate_tags: dict[str, Any],
):
    with pytest.raises(CheckError):
        with instance_cm({"propagate_tags": propagate_tags}) as instance:
            run = instance.create_run_for_job(
                job,
                remote_job_origin=remote_job.get_remote_origin(),
                job_code_origin=remote_job.get_python_origin(),
            )
            instance.launch_run(run.run_id, workspace)


def test_launch_run_with_container_context(
    ecs,
    instance,
    launch_run_with_container_context,
    container_context_config,
):
    existing_tasks = ecs.list_tasks()["taskArns"]

    launch_run_with_container_context(instance)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(existing_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert any(tag == {"key": "HAS_VALUE", "value": "SEE"} for tag in task["tags"])
    assert any(tag == {"key": "DOES_NOT_HAVE_VALUE"} for tag in task["tags"])
    assert any(
        tag == {"key": "ABC", "value": "DEF"} for tag in task["tags"]
    )  # from container context

    assert (
        task.get("overrides").get("memory")
        == container_context_config["ecs"]["run_resources"]["memory"]
    )
    assert (
        task.get("overrides").get("cpu") == container_context_config["ecs"]["run_resources"]["cpu"]
    )
    assert (
        task.get("overrides").get("ephemeralStorage").get("sizeInGiB")
        == container_context_config["ecs"]["run_resources"]["ephemeral_storage"]
    )

    task_definition_arn = task["taskDefinitionArn"]

    task_definition = ecs.describe_task_definition(taskDefinition=task_definition_arn)[
        "taskDefinition"
    ]

    container_definition = task_definition["containerDefinitions"][0]

    assert task_definition["taskRoleArn"] == container_context_config["ecs"]["task_role_arn"]
    assert (
        task_definition["executionRoleArn"] == container_context_config["ecs"]["execution_role_arn"]
    )
    assert task_definition["runtimePlatform"] == container_context_config["ecs"]["runtime_platform"]
    assert task_definition["cpu"] == container_context_config["ecs"]["run_resources"]["cpu"]
    assert task_definition["memory"] == container_context_config["ecs"]["run_resources"]["memory"]
    assert (
        task_definition["ephemeralStorage"]["sizeInGiB"]
        == container_context_config["ecs"]["run_resources"]["ephemeral_storage"]
    )
    assert task_definition["volumes"] == container_context_config["ecs"]["volumes"]

    assert container_definition["mountPoints"] == container_context_config["ecs"]["mount_points"]

    assert (
        container_definition["repositoryCredentials"]["credentialsParameter"]
        == container_context_config["ecs"]["repository_credentials"]
    )

    sidecar_container = task_definition["containerDefinitions"][1]
    assert sidecar_container["name"] == "busyrun"
    assert sidecar_container["image"] == "busybox:latest"


def test_memory_and_cpu(ecs, instance, workspace, run, task_definition):
    # By default
    initial_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(initial_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert task.get("memory") == task_definition.get("memory")
    assert not task.get("overrides").get("memory")
    assert task.get("cpu") == task_definition.get("cpu")
    assert not task.get("overrides").get("cpu")

    # Override memory
    existing_tasks = ecs.list_tasks()["taskArns"]

    instance.add_run_tags(run.run_id, {"ecs/memory": "1024"})
    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(existing_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert task.get("memory") == task_definition.get("memory")
    # taskOverrides expects cpu/memory as strings
    assert task.get("overrides").get("memory") == "1024"
    # taskOverrides expects cpu/memory as integers
    assert task.get("overrides").get("containerOverrides")[0].get("memory") == 1024
    assert task.get("cpu") == task_definition.get("cpu")
    assert not task.get("overrides").get("cpu")
    assert not task.get("overrides").get("containerOverrides")[0].get("cpu")

    # Also override cpu
    existing_tasks = ecs.list_tasks()["taskArns"]

    instance.add_run_tags(run.run_id, {"ecs/cpu": "512"})
    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(existing_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    assert task.get("memory") == task_definition.get("memory")
    assert task.get("overrides").get("memory") == "1024"
    assert task.get("overrides").get("containerOverrides")[0].get("memory") == 1024
    assert task.get("cpu") == task_definition.get("cpu")
    assert task.get("overrides").get("cpu") == "512"
    assert task.get("overrides").get("containerOverrides")[0].get("cpu") == 512

    # Override with invalid constraints
    instance.add_run_tags(run.run_id, {"ecs/memory": "999"})
    with pytest.raises(ClientError):
        instance.launch_run(run.run_id, workspace)


def test_status(
    ecs,
    instance_with_log_group,
    job,
    remote_job,
    cloudwatch_client,
    log_group,
):
    instance = instance_with_log_group

    run = instance.create_run_for_job(
        job,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
        tags={RUN_WORKER_ID_TAG: "abcdef"},
    )

    instance.run_launcher.launch_run(LaunchRunContext(dagster_run=run, workspace=None))

    assert instance.run_launcher.logs == cloudwatch_client

    # Reach into StubbedEcs and grab the task
    # so we can modify its status. This is kind of complicated
    # right now so it might point toward a potential refactor of
    # our internal task data structure - maybe a dict of dicts
    # using cluster and arn as keys - instead of a dict of lists?
    task_arn = instance.get_run_by_id(run.run_id).tags["ecs/task_arn"]
    task = next(task for task in ecs.storage.tasks["default"] if task["taskArn"] == task_arn)

    task_id = task_arn.split("/")[-1]

    for status in RUNNING_STATUSES:
        task["lastStatus"] = status
        running_health_check = instance.run_launcher.check_run_worker_health(run)
        assert running_health_check.status == WorkerStatus.RUNNING

    for status in STOPPED_STATUSES:
        task["lastStatus"] = status

        task["containers"][0]["exitCode"] = 0
        assert instance.run_launcher.check_run_worker_health(run).status == WorkerStatus.SUCCESS

        task["containers"][0]["exitCode"] = 1
        # Without logs (or on a failure fetching logs) the health check sitll fails

        failure_health_check = instance.run_launcher.check_run_worker_health(run)
        assert failure_health_check.status == WorkerStatus.FAILED
        assert (
            failure_health_check.msg
            == f"Task {task_arn} failed.\nStop code: None.\nStop reason: None.\nContainer 'run'"
            " failed - exit code 1.\n"
        )

        assert not failure_health_check.transient
        assert failure_health_check.run_worker_id == "abcdef"

        # with logs, the failure includes the run worker logs

        family = instance.run_launcher._get_run_task_definition_family(run)  # noqa: SLF001
        log_stream = f"{family}/run/{task_id}"
        cloudwatch_client.create_log_stream(logGroupName=log_group, logStreamName=log_stream)

        cloudwatch_client.put_log_events(
            logGroupName=log_group,
            logStreamName=log_stream,
            logEvents=[
                {"timestamp": int(time.time() * 1000), "message": "Oops something bad happened"}
            ],
        )

        failure_health_check = instance.run_launcher.check_run_worker_health(run)
        assert failure_health_check.status == WorkerStatus.FAILED

        assert (
            failure_health_check.msg
            == f"Task {task_arn} failed.\nStop code: None.\nStop reason: None.\nContainer 'run'"
            " failed - exit code 1.\nRun worker logs:\nOops something bad happened"
        )

    task["lastStatus"] = "foo"
    unknown_health_check = instance.run_launcher.check_run_worker_health(run)
    assert unknown_health_check.status == WorkerStatus.UNKNOWN
    assert unknown_health_check.run_worker_id == "abcdef"

    task["lastStatus"] = "STOPPED"
    task["stoppedReason"] = "Timeout waiting for network interface provisioning to complete."

    started_health_check = instance.run_launcher.check_run_worker_health(run)
    assert started_health_check.status == WorkerStatus.FAILED
    assert not started_health_check.transient
    assert started_health_check.run_worker_id == "abcdef"

    task["stoppedReason"] = "Timeout waiting for EphemeralStorage provisioning to complete."

    started_health_check = instance.run_launcher.check_run_worker_health(run)
    assert started_health_check.status == WorkerStatus.FAILED
    assert not started_health_check.transient
    assert started_health_check.run_worker_id == "abcdef"

    task["stoppedReason"] = (
        "CannotPullContainerError: pull image manifest has been retried 5 time(s): Get"
        ' "https://myfakeimage:myfakemanifest":'
        " dial tcp 12.345.678.78:443: i/o timeout"
    )

    started_health_check = instance.run_launcher.check_run_worker_health(run)
    assert started_health_check.status == WorkerStatus.FAILED
    assert not started_health_check.transient
    assert started_health_check.run_worker_id == "abcdef"

    # STARTING runs with these errors are considered a transient failure that can be retried
    starting_run = instance.create_run_for_job(
        job,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
        status=DagsterRunStatus.STARTING,
        tags={RUN_WORKER_ID_TAG: "efghi"},
    )
    instance.run_launcher.launch_run(LaunchRunContext(dagster_run=starting_run, workspace=None))
    task_arn = instance.get_run_by_id(starting_run.run_id).tags["ecs/task_arn"]
    task = next(task for task in ecs.storage.tasks["default"] if task["taskArn"] == task_arn)
    task["lastStatus"] = "STOPPED"
    task["stoppedReason"] = "Timeout waiting for network interface provisioning to complete."

    starting_health_check = instance.run_launcher.check_run_worker_health(starting_run)
    assert starting_health_check.status == WorkerStatus.FAILED
    assert starting_health_check.transient
    assert starting_health_check.run_worker_id == "efghi"


def test_overrides_too_long(
    instance,
    workspace,
    job,
    remote_job,
):
    large_container_context = {i: "boom" for i in range(10000)}

    mock_job_code_origin = JobPythonOrigin(
        job_name="test",
        repository_origin=RepositoryPythonOrigin(
            executable_path="/",
            code_pointer=FileCodePointer(
                python_file="foo.py",
                fn_name="foo",
            ),
            container_image="test:latest",
            container_context=large_container_context,  # pyright: ignore[reportArgumentType]
        ),
    )

    run = instance.create_run_for_job(
        job,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=mock_job_code_origin,
    )

    instance.launch_run(run.run_id, workspace)


def test_custom_launcher(
    ecs,
    custom_instance,
    custom_workspace,
    custom_run,
):
    assert "ecs/task_arn" not in custom_run.tags
    assert "ecs/cluster" not in custom_run.tags

    initial_tasks = ecs.list_tasks()["taskArns"]

    custom_instance.launch_run(custom_run.run_id, custom_workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(initial_tasks)))

    # And we log
    events = custom_instance.event_log_storage.get_logs_for_run(custom_run.run_id)
    latest_event = events[-1]
    assert latest_event.message == "Launching run in custom ECS task"
    metadata = latest_event.dagster_event.engine_event_data.metadata
    assert metadata["Run ID"] == MetadataValue.text(custom_run.run_id)

    # check status and stop task
    assert (
        custom_instance.run_launcher.check_run_worker_health(custom_run).status
        == WorkerStatus.RUNNING
    )
    ecs.stop_task(task=task_arn)


def test_external_launch_type(
    ecs,
    instance_cm,
    workspace,
    remote_job,
    job,
):
    container_name = "external"

    task_definition = ecs.register_task_definition(
        family="external",
        containerDefinitions=[{"name": container_name, "image": "dagster:first"}],
        networkMode="bridge",
        memory="512",
        cpu="256",
    )["taskDefinition"]

    assert task_definition["networkMode"] == "bridge"

    task_definition_arn = task_definition["taskDefinitionArn"]

    # You can provide a family or a task definition ARN
    with instance_cm(
        {
            "task_definition": task_definition_arn,
            "container_name": container_name,
            "run_task_kwargs": {
                "launchType": "EXTERNAL",
            },
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )

        initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
        initial_tasks = ecs.list_tasks()["taskArns"]

        instance.launch_run(run.run_id, workspace)

        # A new task definition is not created
        assert ecs.list_task_definitions()["taskDefinitionArns"] == initial_task_definitions

        # A new task is launched
        tasks = ecs.list_tasks()["taskArns"]

        assert len(tasks) == len(initial_tasks) + 1
        task_arn = next(iter(set(tasks).difference(initial_tasks)))
        task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

        assert task["taskDefinitionArn"] == task_definition["taskDefinitionArn"]
        assert task["launchType"] == "EXTERNAL"


def test_removing_network_configuration(
    ecs,
    instance_cm,
    workspace,
    remote_job,
    job,
):
    container_name = "external"

    task_definition = ecs.register_task_definition(
        family="external",
        containerDefinitions=[{"name": container_name, "image": "dagster:first"}],
        networkMode="bridge",
        memory="512",
        cpu="256",
    )["taskDefinition"]

    assert task_definition["networkMode"] == "bridge"
    task_definition_arn = task_definition["taskDefinitionArn"]

    # You can provide a family or a task definition ARN
    with instance_cm(
        {
            "task_definition": task_definition_arn,
            "container_name": container_name,
            "run_task_kwargs": {"networkConfiguration": None},
        }
    ) as instance:
        run = instance.create_run_for_job(
            job,
            remote_job_origin=remote_job.get_remote_origin(),
            job_code_origin=remote_job.get_python_origin(),
        )

        initial_task_definitions = ecs.list_task_definitions()["taskDefinitionArns"]
        initial_tasks = ecs.list_tasks()["taskArns"]

        instance.launch_run(run.run_id, workspace)

        # A new task definition is not created
        assert ecs.list_task_definitions()["taskDefinitionArns"] == initial_task_definitions

        # A new task is launched
        tasks = ecs.list_tasks()["taskArns"]

        assert len(tasks) == len(initial_tasks) + 1
        task_arn = next(iter(set(tasks).difference(initial_tasks)))
        task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

        assert task["taskDefinitionArn"] == task_definition["taskDefinitionArn"]
        assert "networkConfiguration" not in task


def test_container_overrides(ecs, instance, workspace, run, task_definition):
    # By default, no resource requirements
    initial_tasks = ecs.list_tasks()["taskArns"]

    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(initial_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    container_override = task.get("overrides").get("containerOverrides")[0]
    assert container_override.get("name") == "run"
    assert not container_override.get("resourceRequirements")

    # Override with resource requirements (e.g., GPU)
    existing_tasks = ecs.list_tasks()["taskArns"]

    container_overrides = {
        "resourceRequirements": [{"type": "GPU", "value": "1"}],
    }
    instance.add_run_tags(run.run_id, {"ecs/container_overrides": json.dumps(container_overrides)})
    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(existing_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    container_override = task.get("overrides").get("containerOverrides")[0]
    assert container_override.get("name") == "run"
    assert container_override.get("resourceRequirements") == [{"type": "GPU", "value": "1"}]

    # Override with multiple fields
    existing_tasks = ecs.list_tasks()["taskArns"]

    container_overrides = {
        "resourceRequirements": [{"type": "GPU", "value": "2"}],
        "environment": [{"name": "CUSTOM_VAR", "value": "custom_value"}],
    }
    instance.add_run_tags(run.run_id, {"ecs/container_overrides": json.dumps(container_overrides)})
    instance.launch_run(run.run_id, workspace)

    tasks = ecs.list_tasks()["taskArns"]
    task_arn = next(iter(set(tasks).difference(existing_tasks)))
    task = ecs.describe_tasks(tasks=[task_arn])["tasks"][0]

    container_override = task.get("overrides").get("containerOverrides")[0]
    assert container_override.get("name") == "run"
    assert container_override.get("resourceRequirements") == [{"type": "GPU", "value": "2"}]
    assert container_override.get("environment") == [
        {"name": "CUSTOM_VAR", "value": "custom_value"}
    ]



import json
import re
import sys
import threading
import time
from datetime import datetime
from threading import Event
from typing import TYPE_CHECKING
from uuid import uuid4

import pytest
from dagster import asset, materialize, open_pipes_session
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckKey, AssetCheckSpec
from dagster._core.definitions.data_version import (
    DATA_VERSION_IS_USER_PROVIDED_TAG,
    DATA_VERSION_TAG,
)
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MarkdownMetadataValue
from dagster._core.execution.context.compute import AssetExecutionContext
from dagster._core.instance_for_test import instance_for_test
from dagster._core.pipes.subprocess import PipesSubprocessClient
from dagster._core.pipes.utils import PipesEnvContextInjector
from dagster._core.storage.asset_check_execution_record import AssetCheckExecutionRecordStatus
from dagster_pipes import _make_message

from dagster_aws.pipes import (
    PipesCloudWatchLogReader,
    PipesCloudWatchMessageReader,
    PipesS3ContextInjector,
    PipesS3LogReader,
    PipesS3MessageReader,
)
from dagster_aws_tests.pipes_tests.utils import _PYTHON_EXECUTABLE, _S3_TEST_BUCKET

if TYPE_CHECKING:
    from mypy_boto3_logs import CloudWatchLogsClient


def test_s3_log_reader(s3_client, capsys):
    key = str(uuid4())
    log_reader = PipesS3LogReader(client=s3_client, bucket=_S3_TEST_BUCKET, key=key)
    is_session_closed = Event()

    assert not log_reader.target_is_readable({})

    s3_client.put_object(Bucket=_S3_TEST_BUCKET, Key=key, Body=b"Line 0\nLine 1")

    assert log_reader.target_is_readable({})

    log_reader.start({}, is_session_closed)
    assert log_reader.is_running()

    s3_client.put_object(Bucket=_S3_TEST_BUCKET, Key=key, Body=b"Line 0\nLine 1\nLine 2")

    is_session_closed.set()

    log_reader.stop()

    assert not log_reader.is_running()

    captured = capsys.readouterr()

    assert captured.out == "Line 0\nLine 1\nLine 2"

    assert sys.stdout is not None


def test_s3_pipes_components(
    capsys,
    tmpdir,
    external_script,
    s3_client,
):
    context_injector = PipesS3ContextInjector(bucket=_S3_TEST_BUCKET, client=s3_client)
    message_reader = PipesS3MessageReader(bucket=_S3_TEST_BUCKET, client=s3_client, interval=0.001)

    @asset(check_specs=[AssetCheckSpec(name="foo_check", asset=AssetKey(["foo"]))])
    def foo(context: AssetExecutionContext, ext: PipesSubprocessClient):
        return ext.run(
            command=[_PYTHON_EXECUTABLE, external_script],
            context=context,
            extras={"bar": "baz"},
            env={},
        ).get_results()

    resource = PipesSubprocessClient(
        context_injector=context_injector, message_reader=message_reader
    )

    with instance_for_test() as instance:
        materialize([foo], instance=instance, resources={"ext": resource})
        mat = instance.get_latest_materialization_event(foo.key)
        assert mat and mat.asset_materialization
        assert isinstance(mat.asset_materialization.metadata["bar"], MarkdownMetadataValue)
        assert mat.asset_materialization.metadata["bar"].value == "baz"
        assert mat.asset_materialization.tags
        assert mat.asset_materialization.tags[DATA_VERSION_TAG] == "alpha"
        assert mat.asset_materialization.tags[DATA_VERSION_IS_USER_PROVIDED_TAG]

        captured = capsys.readouterr()
        assert re.search(r"dagster - INFO - [^\n]+ - hello world\n", captured.err, re.MULTILINE)

        asset_check_executions = instance.event_log_storage.get_asset_check_execution_history(
            check_key=AssetCheckKey(foo.key, name="foo_check"),
            limit=1,
        )
        assert len(asset_check_executions) == 1
        assert asset_check_executions[0].status == AssetCheckExecutionRecordStatus.SUCCEEDED


def test_s3_context_injector(s3_client):
    context_injector = PipesS3ContextInjector(
        bucket=_S3_TEST_BUCKET, client=s3_client, key_prefix="custom_prefix/"
    )
    pipes_context = {
        "asset_keys": None,
        "code_version_by_asset_key": None,
        "extras": {},
        "job_name": None,
        "partition_key": None,
        "partition_key_range": None,
        "partition_time_window": None,
        "provenance_by_asset_key": None,
        "retry_number": 0,
        "run_id": "test-run-id",
    }

    with context_injector.inject_context(pipes_context) as context:  # type: ignore
        context_key = context["key"]
        assert context["bucket"] == _S3_TEST_BUCKET
        assert context_key.startswith("custom_prefix/")
        assert (
            json.loads(
                s3_client.get_object(Bucket=context["bucket"], Key=context_key)["Body"]
                .read()
                .decode("utf-8")
            )
            == pipes_context
        )

    # Make sure the object has been deleted
    with pytest.raises(
        s3_client.exceptions.NoSuchKey,
        match="The specified key does not exist",
    ):
        s3_client.get_object(Bucket=_S3_TEST_BUCKET, Key=context_key)


def test_cloudwatch_logs_reader(cloudwatch_client: "CloudWatchLogsClient", capsys):
    reader = PipesCloudWatchLogReader(client=cloudwatch_client)

    is_session_closed = threading.Event()

    log_group = "/pipes/tests"
    log_stream = "test-cloudwatch-logs-reader"

    assert not reader.target_is_readable({"log_group": log_group, "log_stream": log_stream}), (
        "Should not be able to read without the stream existing"
    )

    cloudwatch_client.create_log_group(logGroupName=log_group)
    cloudwatch_client.create_log_stream(logGroupName=log_group, logStreamName=log_stream)

    cloudwatch_client.put_log_events(
        logGroupName=log_group,
        logStreamName=log_stream,
        logEvents=[{"timestamp": int(datetime.now().timestamp() * 1000), "message": "1"}],
    )

    time.sleep(0.1)

    assert reader.target_is_readable({"log_group": log_group, "log_stream": log_stream}), (
        "Should be able to read after the stream is created"
    )

    reader.start({"log_group": log_group, "log_stream": log_stream}, is_session_closed)

    cloudwatch_client.put_log_events(
        logGroupName=log_group,
        logStreamName=log_stream,
        logEvents=[
            {
                "timestamp": int(datetime.now().timestamp() * 1000),
                "message": "2",
            }
        ],
    )

    time.sleep(0.1)

    cloudwatch_client.put_log_events(
        logGroupName=log_group,
        logStreamName=log_stream,
        logEvents=[
            {
                "timestamp": int(datetime.now().timestamp() * 1000),
                "message": "3",
            }
        ],
    )

    time.sleep(0.1)

    is_session_closed.set()

    time.sleep(1)

    reader.stop()

    time.sleep(1)

    assert capsys.readouterr().out.strip() == "1\n2\n3"


def test_cloudwatch_message_reader(cloudwatch_client: "CloudWatchLogsClient", capsys):
    log_group = "/pipes/tests/messages"
    messages_log_stream = "test-cloudwatch-messages-reader"
    logs_log_stream = "test-cloudwatch-messages-reader-stdout"

    reader = PipesCloudWatchMessageReader(client=cloudwatch_client)

    @asset
    def my_asset(context: AssetExecutionContext):
        with open_pipes_session(
            context=context, message_reader=reader, context_injector=PipesEnvContextInjector()
        ) as session:
            assert not reader.messages_are_readable({})

            cloudwatch_client.create_log_group(logGroupName=log_group)
            cloudwatch_client.create_log_stream(
                logGroupName=log_group, logStreamName=messages_log_stream
            )
            cloudwatch_client.create_log_stream(
                logGroupName=log_group, logStreamName=logs_log_stream
            )

            assert not reader.messages_are_readable({})

            new_params = {
                "log_group": log_group,
                "log_stream": messages_log_stream,
            }

            session.report_launched({"extras": new_params})

            assert reader.launched_payload is not None

            assert reader.messages_are_readable(reader.launched_payload["extras"])

            def log_event(message: str):
                cloudwatch_client.put_log_events(
                    logGroupName=log_group,
                    logStreamName=messages_log_stream,
                    logEvents=[
                        {"timestamp": int(datetime.now().timestamp() * 1000), "message": message}
                    ],
                )

            def log_line(message: str):
                cloudwatch_client.put_log_events(
                    logGroupName=log_group,
                    logStreamName=logs_log_stream,
                    logEvents=[
                        {"timestamp": int(datetime.now().timestamp() * 1000), "message": message}
                    ],
                )

            log_event(json.dumps(_make_message(method="opened", params={})))

            log_event(
                json.dumps(
                    _make_message(method="log", params={"message": "Hello!", "level": "INFO"})
                )
            )

            log_event(
                json.dumps(
                    _make_message(
                        method="report_asset_materialization",
                        params={
                            "asset_key": "my_asset",
                            "metadata": {"foo": {"raw_value": "bar", "type": "text"}},
                            "data_version": "alpha",
                        },
                    )
                )
            )

            log_event(json.dumps(_make_message(method="closed", params={})))

            log_line("Hello 1")

            log_reader = PipesCloudWatchLogReader(
                client=cloudwatch_client,
                log_group=log_group,
                log_stream=logs_log_stream,
                target_stream=sys.stdout,
                start_time=int(session.created_at.timestamp() * 1000),
            )
            reader.add_log_reader(log_reader)

            log_line("Hello 2")

            assert log_reader.target_is_readable({})

        return session.get_results()

    result = materialize([my_asset])

    assert result.success

    mats = result.get_asset_materialization_events()
    assert len(mats) == 1
    mat = mats[0]
    assert mat.asset_key == AssetKey(["my_asset"])
    assert mat.materialization.metadata["foo"].value == "bar"
    assert mat.materialization.tags[DATA_VERSION_TAG] == "alpha"  # type: ignore

    captured = capsys.readouterr()
    assert "Hello 1" in captured.out
    assert "Hello 2" in captured.out



from typing import TYPE_CHECKING
from unittest.mock import MagicMock
from uuid import uuid4

import boto3
import pytest
import pytest_cases
from dagster import asset, materialize, open_pipes_session
from dagster._core.execution.context.compute import AssetExecutionContext
from dagster._core.instance_for_test import instance_for_test

from dagster_aws.pipes import PipesEMRContainersClient, PipesS3MessageReader
from dagster_aws_tests.pipes_tests.utils import _MOTO_SERVER_URL, _S3_TEST_BUCKET

if TYPE_CHECKING:
    from mypy_boto3_emr_containers import EMRContainersClient


@pytest.fixture
def emr_containers_client(moto_server) -> "EMRContainersClient":
    return boto3.client("emr-containers", region_name="us-east-1", endpoint_url=_MOTO_SERVER_URL)


@pytest_cases.fixture
@pytest_cases.parametrize(pipes_params_bootstrap_method=["args", "env"])
def pipes_emr_containers_client(
    emr_containers_client, s3_client, pipes_params_bootstrap_method
) -> PipesEMRContainersClient:
    return PipesEMRContainersClient(
        client=emr_containers_client,
        message_reader=PipesS3MessageReader(bucket=_S3_TEST_BUCKET, client=s3_client),
        pipes_params_bootstrap_method=pipes_params_bootstrap_method,
    )


@asset
def manual_asset(
    context: AssetExecutionContext, pipes_emr_containers_client: PipesEMRContainersClient
):
    with open_pipes_session(
        context=context,
        message_reader=pipes_emr_containers_client.message_reader,
        context_injector=pipes_emr_containers_client.context_injector,
    ) as session:
        params = pipes_emr_containers_client._enrich_start_params(  # noqa: SLF001
            context=context,
            session=session,
            params={
                "virtualClusterId": "test",
                "executionRoleArn": "arn:aws:iam::123456789012:role/EMRServerlessRole",
                "jobDriver": {
                    "sparkSubmitJobDriver": {
                        "entryPoint": "s3://my-bucket/my-script.py",
                        "entryPointArguments": ["arg1", "arg2"],
                        "sparkSubmitParameters": "--conf spark.executor.instances=2",
                    }
                },
                "clientToken": str(uuid4()),
            },
        )
        assert params.get("tags", {}).get("dagster/run-id") == context.run_id

        pipes_emr_containers_client._start(context, params)  # noqa: SLF001

        pipes_emr_containers_client._client.start_job_run.assert_called_once()  # noqa: SLF001  # pyright: ignore[reportFunctionMemberAccess]
        assert (
            pipes_emr_containers_client._client.start_job_run.call_args.kwargs["virtualClusterId"]  # noqa: SLF001  # pyright: ignore[reportFunctionMemberAccess]
            == "test"
        )

        if pipes_emr_containers_client.pipes_params_bootstrap_method == "args":
            assert (
                "--dagster-pipes-context"
                in pipes_emr_containers_client._client.start_job_run.call_args.kwargs["jobDriver"][  # noqa: SLF001  # pyright: ignore[reportFunctionMemberAccess]
                    "sparkSubmitJobDriver"
                ]["sparkSubmitParameters"]
            )

            assert (
                "--dagster-pipes-messages"
                in pipes_emr_containers_client._client.start_job_run.call_args.kwargs["jobDriver"][  # noqa: SLF001  # pyright: ignore[reportFunctionMemberAccess]
                    "sparkSubmitJobDriver"
                ]["sparkSubmitParameters"]
            )

        elif pipes_emr_containers_client.pipes_params_bootstrap_method == "env":
            configurations = pipes_emr_containers_client._client.start_job_run.call_args.kwargs[  # noqa: SLF001  # pyright: ignore[reportFunctionMemberAccess]
                "configurationOverrides"
            ]["applicationConfiguration"]

            spark_defaults_injected = False
            for configuration in configurations:
                if configuration.get("classification") == "spark-defaults":
                    if (
                        "spark.yarn.appMasterEnv.DAGSTER_PIPES_CONTEXT"
                        in configuration["properties"]
                        and "spark.yarn.appMasterEnv.DAGSTER_PIPES_MESSAGES"
                        in configuration["properties"]
                    ):
                        spark_defaults_injected = True

            assert spark_defaults_injected

        return session.get_results()


def test_emr_containers_manual(pipes_emr_containers_client: "PipesEMRContainersClient"):
    pipes_emr_containers_client._client = MagicMock()  # noqa: SLF001
    with instance_for_test() as instance:
        materialize(
            [manual_asset],
            resources={"pipes_emr_containers_client": pipes_emr_containers_client},
            instance=instance,
        )


@asset
def emr_containers_asset(
    context: AssetExecutionContext, pipes_emr_containers_client: PipesEMRContainersClient
):
    return pipes_emr_containers_client.run(
        context=context,
        extras={"bar": "baz"},
        start_job_run_params={
            "virtualClusterId": "test",
            "executionRoleArn": "arn:aws:iam::123456789012:role/EMRServerlessRole",
            "jobDriver": {
                "sparkSubmitJobDriver": {
                    "entryPoint": "s3://my-bucket/my-script.py",
                    "entryPointArguments": ["arg1", "arg2"],
                    "sparkSubmitParameters": "--conf spark.executor.instances=2",
                }
            },
            "clientToken": str(uuid4()),
        },
    ).get_materialize_result()



import base64
import json

import pytest
from dagster import asset, materialize, open_pipes_session
from dagster._core.pipes.utils import PipesEnvContextInjector

from dagster_aws.pipes import PipesLambdaClient, PipesLambdaLogsMessageReader, PipesS3MessageReader
from dagster_aws_tests.pipes_tests.fake_lambda import (
    LOG_TAIL_LIMIT,
    FakeLambdaClient,
    LambdaFunctions,
)
from dagster_aws_tests.pipes_tests.utils import _S3_TEST_BUCKET


def test_fake_lambda_logs():
    event = {}
    response = FakeLambdaClient().invoke(
        FunctionName=LambdaFunctions.trunc_logs.__name__,
        InvocationType="RequestResponse",
        Payload=json.dumps(event),
        LogType="Tail",
    )

    log_result = base64.b64decode(response["LogResult"])
    assert len(log_result) == LOG_TAIL_LIMIT

    small_size = 512
    response = FakeLambdaClient().invoke(
        FunctionName=LambdaFunctions.small_logs.__name__,
        InvocationType="RequestResponse",
        Payload=json.dumps({"size": small_size}),
        LogType="Tail",
    )

    log_result = base64.b64decode(response["LogResult"])
    assert len(log_result) == small_size + 1  # size + \n


def test_manual_fake_lambda_pipes():
    @asset
    def fake_lambda_asset(context):
        context_injector = PipesEnvContextInjector()
        message_reader = PipesLambdaLogsMessageReader()

        with open_pipes_session(
            context=context,
            message_reader=message_reader,
            context_injector=context_injector,
        ) as session:
            user_event = {}
            response = FakeLambdaClient().invoke(
                FunctionName=LambdaFunctions.pipes_basic.__name__,
                InvocationType="RequestResponse",
                Payload=json.dumps(
                    {
                        **user_event,
                        **session.get_bootstrap_env_vars(),
                    }
                ),
                LogType="Tail",
            )
            message_reader.consume_lambda_logs(response)
            yield from session.get_results()

    result = materialize([fake_lambda_asset])
    assert result.success
    mat_evts = result.get_asset_materialization_events()
    assert len(mat_evts) == 1
    assert mat_evts[0].materialization.metadata["meta"].value == "data"


@pytest.mark.parametrize(
    "lambda_fn",
    [
        LambdaFunctions.pipes_basic.__name__,
        LambdaFunctions.pipes_messy_logs.__name__,
    ],
)
def test_client_lambda_pipes(lambda_fn):
    @asset
    def fake_lambda_asset(context):
        return (
            PipesLambdaClient(FakeLambdaClient())
            .run(
                context=context,
                function_name=lambda_fn,
                event={},
            )
            .get_materialize_result()
        )

    result = materialize([fake_lambda_asset])
    assert result.success
    mat_evts = result.get_asset_materialization_events()
    assert len(mat_evts) == 1
    assert mat_evts[0].materialization.metadata["meta"].value == "data"


def test_fake_client_lambda_error():
    @asset
    def fake_lambda_asset(context):
        yield from (
            PipesLambdaClient(FakeLambdaClient())
            .run(
                context=context,
                function_name=LambdaFunctions.error.__name__,
                event={},
            )
            .get_results()
        )

    with pytest.raises(Exception, match="Lambda Function Error"):
        materialize([fake_lambda_asset])


def test_lambda_s3_pipes(s3_client):
    @asset
    def fake_lambda_asset(context):
        return (
            PipesLambdaClient(
                FakeLambdaClient(),
                message_reader=PipesS3MessageReader(
                    client=s3_client,
                    bucket=_S3_TEST_BUCKET,
                    interval=0.01,
                ),
            )
            .run(
                context=context,
                function_name=LambdaFunctions.pipes_s3_messages.__name__,
                event={},
            )
            .get_materialize_result()
        )

    result = materialize([fake_lambda_asset])
    assert result.success
    mat_evts = result.get_asset_materialization_events()
    assert len(mat_evts) == 1
    assert mat_evts[0].materialization.metadata["meta"].value == "data"



import dagster


@dagster.op
def node(_):
    pass


@dagster.job
def job():
    node()


@dagster.repository  # pyright: ignore[reportArgumentType]
def repository():
    return {"jobs": {"job": job}}



import inspect
import shutil
import textwrap
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from typing import Any

_PYTHON_EXECUTABLE = shutil.which("python") or "python"
_S3_TEST_BUCKET = "pipes-testing"
_MOTO_SERVER_PORT = 5193
_MOTO_SERVER_URL = f"http://localhost:{_MOTO_SERVER_PORT}"


@contextmanager
def temp_script(script_fn: Callable[[], Any]) -> Iterator[str]:
    # drop the signature line
    source = textwrap.dedent(inspect.getsource(script_fn).split("\n", 1)[1])
    with NamedTemporaryFile() as file:
        file.write(source.encode())
        file.flush()
        yield file.name



from dagster_aws.ssm.parameters import get_parameters_by_paths, get_parameters_by_tags


def test_get_parameters_from_tags(mock_ssm_client):
    mock_ssm_client.put_parameter(
        Name="foo_param1",
        Type="String",
        Value="foo_value1",
        Tags=[{"Key": "foo_tag_key", "Value": "foo_tag_value1"}],
    )
    mock_ssm_client.put_parameter(
        Name="foo_param2",
        Type="String",
        Value="foo_value2",
        Tags=[{"Key": "foo_tag_key", "Value": "foo_tag_value2"}],
    )
    mock_ssm_client.put_parameter(
        Name="bar_param",
        Type="String",
        Value="bar_value",
        Tags=[{"Key": "bar_tag_key", "Value": "bar_tag_value"}],
    )
    result = get_parameters_by_tags(
        mock_ssm_client,
        parameter_tags=[{"key": "foo_tag_key", "values": ["foo_tag_value1", "foo_tag_value2"]}],
        with_decryption=False,
    )
    assert result == {"foo_param1": "foo_value1", "foo_param2": "foo_value2"}

    result = get_parameters_by_tags(
        mock_ssm_client,
        parameter_tags=[{"key": "non-existent", "values": ["non-existent"]}],
        with_decryption=False,
    )
    assert result == {}


def test_get_parameters_by_path(mock_ssm_client):
    mock_ssm_client.put_parameter(Name="path/based/param1", Value="param1", Type="String")
    mock_ssm_client.put_parameter(Name="path/based/nested/param2", Value="param2", Type="String")
    mock_ssm_client.put_parameter(Name="path/not/matching/param3", Value="param3", Type="String")

    result = get_parameters_by_paths(
        mock_ssm_client,
        ["path/based"],  # pyright: ignore[reportArgumentType]
        with_decryption=True,
        recursive=True,
    )
    assert result == {"path/based/param1": "param1", "path/based/nested/param2": "param2"}

    result = get_parameters_by_paths(
        mock_ssm_client,
        ["path/based"],  # pyright: ignore[reportArgumentType]
        with_decryption=True,
        recursive=False,
    )
    assert result == {"path/based/param1": "param1"}

    result = get_parameters_by_paths(
        mock_ssm_client,
        ["path/"],  # pyright: ignore[reportArgumentType]
        with_decryption=False,
        recursive=False,
    )
    assert result == {}



import json
from typing import Callable, ContextManager  # noqa: UP035

from dagster import job, op, repository
from dagster._config import process_config, resolve_to_config_type
from dagster._core.definitions.reconstruct import reconstructable
from dagster._core.execution.api import create_execution_plan
from dagster._core.execution.context.system import PlanData, PlanOrchestrationContext
from dagster._core.execution.context_creation_job import create_context_free_log_manager
from dagster._core.execution.retries import RetryMode
from dagster._core.executor.init import InitExecutorContext
from dagster._core.executor.step_delegating.step_handler.base import StepHandlerContext
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation.handle import RepositoryHandle
from dagster._core.storage.fs_io_manager import fs_io_manager
from dagster._core.test_utils import create_run_for_test, in_process_test_workspace
from dagster._core.types.loadable_target_origin import LoadableTargetOrigin
from dagster._grpc.types import ExecuteStepArgs
from dagster._utils.hosted_user_process import remote_job_from_recon_job

from dagster_aws.ecs.executor import _ECS_EXECUTOR_CONFIG_SCHEMA, ecs_executor


@job(
    executor_def=ecs_executor,
    resource_defs={"io_manager": fs_io_manager},
)
def bar():
    @op(
        tags={
            "ecs/cpu": "1024",
            "ecs/memory": "512",
        }
    )
    def foo():
        return 1

    foo()


@repository
def bar_repo():
    return [bar]


def _get_executor(instance, job_def, executor_config=None):
    process_result = process_config(
        resolve_to_config_type(_ECS_EXECUTOR_CONFIG_SCHEMA), executor_config or {}
    )
    if not process_result.success:
        raise AssertionError(f"Process result errors: {process_result.errors}")

    return ecs_executor.executor_creation_fn(  # type: ignore
        InitExecutorContext(
            job=job_def,
            executor_def=ecs_executor,
            executor_config=process_result.value,  # type: ignore
            instance=instance,
        )
    )


def _step_handler_context(job_def, dagster_run, instance, executor):
    execution_plan = create_execution_plan(job_def)
    log_manager = create_context_free_log_manager(instance, dagster_run)

    plan_context = PlanOrchestrationContext(
        plan_data=PlanData(
            job=job_def,
            dagster_run=dagster_run,
            instance=instance,
            execution_plan=execution_plan,
            raise_on_error=True,
            retry_mode=RetryMode.DISABLED,
        ),
        log_manager=log_manager,
        executor=executor,
        output_capture=None,
    )

    execute_step_args = ExecuteStepArgs(
        reconstructable(bar).get_python_origin(),
        dagster_run.run_id,
        ["foo"],
        print_serialized_events=False,
    )

    return StepHandlerContext(
        instance=instance,
        plan_context=plan_context,
        steps=execution_plan.steps,  # type: ignore
        execute_step_args=execute_step_args,
    )


def test_executor_init(instance_cm: Callable[..., ContextManager[DagsterInstance]]):
    with instance_cm() as instance:
        recon_job = reconstructable(bar)
        loadable_target_origin = LoadableTargetOrigin(python_file=__file__, attribute="bar_repo")

        memory = 128
        cpu = 500
        env_var = {"key": "OVERRIDE_VAR", "value": "foo"}
        executor = _get_executor(
            instance,
            reconstructable(bar),
            {
                "cpu": cpu,
                "memory": memory,
                "task_overrides": {
                    "containerOverrides": [
                        {
                            "name": "run",
                            "environment": [env_var],
                        }
                    ],
                },
            },
        )

        with in_process_test_workspace(
            instance, loadable_target_origin, container_image="testing/dagster"
        ) as workspace:
            location = workspace.get_code_location(workspace.code_location_names[0])
            repo_handle = RepositoryHandle.from_location(
                repository_name="bar_repo",
                code_location=location,
            )
            fake_remote_job = remote_job_from_recon_job(
                recon_job,
                op_selection=None,
                repository_handle=repo_handle,
            )

            run = create_run_for_test(
                instance,
                job_name="bar",
                remote_job_origin=fake_remote_job.get_remote_origin(),
                job_code_origin=recon_job.get_python_origin(),
            )
            step_handler_context = _step_handler_context(
                job_def=reconstructable(bar),
                dagster_run=run,
                instance=instance,
                executor=executor,
            )
            run_task_kwargs = executor._step_handler._get_run_task_kwargs(  # type: ignore  # noqa: SLF001
                run,
                ["my-command"],
                "asdasd",
                {},
                step_handler_context,
                executor._step_handler._get_container_context(step_handler_context),  # type: ignore  # noqa: SLF001
            )

            assert run_task_kwargs["launchType"] == "FARGATE"  # this comes from the Run Launcher

            overrides = run_task_kwargs["overrides"]

            assert overrides["cpu"] == str(cpu)
            assert overrides["memory"] == str(memory)

            run_container_overrides = overrides["containerOverrides"][0]

            assert run_container_overrides["name"] == "run"
            assert run_container_overrides["command"] == ["my-command"]

            assert env_var in run_container_overrides["environment"]


def test_executor_sanitized_step_key(instance_cm: Callable[..., ContextManager[DagsterInstance]]):
    with instance_cm() as instance:
        recon_job = reconstructable(bar)
        loadable_target_origin = LoadableTargetOrigin(python_file=__file__, attribute="bar_repo")

        executor = _get_executor(
            instance,
            reconstructable(bar),
        )

        with in_process_test_workspace(
            instance, loadable_target_origin, container_image="testing/dagster"
        ) as workspace:
            location = workspace.get_code_location(workspace.code_location_names[0])
            repo_handle = RepositoryHandle.from_location(
                repository_name="bar_repo",
                code_location=location,
            )
            fake_remote_job = remote_job_from_recon_job(
                recon_job,
                op_selection=None,
                repository_handle=repo_handle,
            )

            run = create_run_for_test(
                instance,
                job_name="bar",
                remote_job_origin=fake_remote_job.get_remote_origin(),
                job_code_origin=recon_job.get_python_origin(),
            )
            step_handler_context = _step_handler_context(
                job_def=reconstructable(bar),
                dagster_run=run,
                instance=instance,
                executor=executor,
            )
            run_task_kwargs = executor._step_handler._get_run_task_kwargs(  # type: ignore  # noqa: SLF001
                run,
                ["my-command"],
                "foo.bar[filename_0]",
                {},
                step_handler_context,
                executor._step_handler._get_container_context(step_handler_context),  # type: ignore  # noqa: SLF001
            )

            tags = {tag["key"]: tag["value"] for tag in run_task_kwargs["tags"]}
            assert tags["dagster/step-key"] == "foo-bar-filename_0"


def test_executor_launch(instance_cm: Callable[..., ContextManager[DagsterInstance]]):
    with instance_cm() as instance:
        recon_job = reconstructable(bar)
        loadable_target_origin = LoadableTargetOrigin(python_file=__file__, attribute="bar_repo")

        with in_process_test_workspace(
            instance, loadable_target_origin, container_image="testing/dagster"
        ) as workspace:
            location = workspace.get_code_location(workspace.code_location_names[0])
            repo_handle = RepositoryHandle.from_location(
                repository_name="bar_repo",
                code_location=location,
            )
            fake_remote_job = remote_job_from_recon_job(
                recon_job,
                op_selection=None,
                repository_handle=repo_handle,
            )

            executor = _get_executor(instance, reconstructable(bar), {})
            run = create_run_for_test(
                instance,
                job_name="bar",
                remote_job_origin=fake_remote_job.get_remote_origin(),
                job_code_origin=recon_job.get_python_origin(),
            )
            step_handler_context = _step_handler_context(
                job_def=reconstructable(bar),
                dagster_run=run,
                instance=instance,
                executor=executor,
            )
            from unittest.mock import MagicMock

            executor._step_handler.ecs.run_task = MagicMock(  # type: ignore  # noqa: SLF001
                return_value={"tasks": [{"taskArn": "arn:123"}]}
            )

            next(iter(executor._step_handler.launch_step(step_handler_context)))  # type: ignore  # noqa: SLF001

            run_task_kwargs = executor._step_handler.ecs.run_task.call_args[1]  # type: ignore  # noqa: SLF001

            # resources should come from step tags
            assert run_task_kwargs["overrides"]["cpu"] == "1024"
            assert run_task_kwargs["overrides"]["memory"] == "512"

            tags = run_task_kwargs["tags"]

            assert {
                "key": "dagster/run-id",
                "value": run.run_id,
            } in tags

            assert {
                "key": "dagster/job",
                "value": run.job_name,
            } in tags

            assert {
                "key": "dagster/step-key",
                "value": "foo",
            } in tags

            assert run_task_kwargs["overrides"]["containerOverrides"][0]["command"] == [
                "dagster",
                "api",
                "execute_step",
            ]

            found_executor_args_var = False
            for var in run_task_kwargs["overrides"]["containerOverrides"][0]["environment"]:
                if var["name"] == "DAGSTER_COMPRESSED_EXECUTE_STEP_ARGS":
                    found_executor_args_var = True
                    break

            assert found_executor_args_var


def test_executor_container_overrides(instance_cm: Callable[..., ContextManager[DagsterInstance]]):
    with instance_cm() as instance:
        recon_job = reconstructable(bar)
        loadable_target_origin = LoadableTargetOrigin(python_file=__file__, attribute="bar_repo")

        executor = _get_executor(
            instance,
            reconstructable(bar),
        )

        with in_process_test_workspace(
            instance, loadable_target_origin, container_image="testing/dagster"
        ) as workspace:
            location = workspace.get_code_location(workspace.code_location_names[0])
            repo_handle = RepositoryHandle.from_location(
                repository_name="bar_repo",
                code_location=location,
            )
            fake_remote_job = remote_job_from_recon_job(
                recon_job,
                op_selection=None,
                repository_handle=repo_handle,
            )

            run = create_run_for_test(
                instance,
                job_name="bar",
                remote_job_origin=fake_remote_job.get_remote_origin(),
                job_code_origin=recon_job.get_python_origin(),
            )
            step_handler_context = _step_handler_context(
                job_def=reconstructable(bar),
                dagster_run=run,
                instance=instance,
                executor=executor,
            )

            # Test with ecs/container_overrides step tag
            container_overrides = {
                "resourceRequirements": [{"type": "GPU", "value": "1"}],
            }
            step_tags = {"ecs/container_overrides": json.dumps(container_overrides)}

            run_task_kwargs = executor._step_handler._get_run_task_kwargs(  # type: ignore  # noqa: SLF001
                run,
                ["my-command"],
                "test_step",
                step_tags,
                step_handler_context,
                executor._step_handler._get_container_context(step_handler_context),  # type: ignore  # noqa: SLF001
            )

            run_container_overrides = run_task_kwargs["overrides"]["containerOverrides"][0]

            assert run_container_overrides["name"] == "run"
            assert run_container_overrides["command"] == ["my-command"]
            assert run_container_overrides["resourceRequirements"] == [
                {"type": "GPU", "value": "1"}
            ]



import os
from unittest import mock

import pytest
from dagster import DagsterEvent, EventLogEntry, build_init_resource_context
from dagster._core.execution.plan.objects import StepSuccessData

from dagster_aws.emr.pyspark_step_launcher import EmrPySparkStepLauncher, emr_pyspark_step_launcher

EVENTS = [
    EventLogEntry(
        run_id="1",
        error_info=None,
        level=20,
        user_message="foo",
        timestamp=1.0,
        dagster_event=DagsterEvent(event_type_value="STEP_START", job_name="foo"),
    ),
    EventLogEntry(
        run_id="1",
        error_info=None,
        level=20,
        user_message="bar",
        timestamp=2.0,
        dagster_event=None,
    ),
    EventLogEntry(
        run_id="1",
        error_info=None,
        level=20,
        user_message="baz",
        timestamp=3.0,
        dagster_event=DagsterEvent(
            event_type_value="STEP_SUCCESS",
            job_name="foo",
            event_specific_data=StepSuccessData(duration_ms=2.0),
        ),
    ),
]


@mock.patch(
    "dagster_aws.emr.emr.EmrJobRunner.is_emr_step_complete", side_effect=[False, False, True]
)
@mock.patch(
    "dagster_aws.emr.pyspark_step_launcher.EmrPySparkStepLauncher.read_events",
    side_effect=[EVENTS[0:1], [], EVENTS[0:3]],
)
def test_wait_for_completion(_mock_is_emr_step_complete, _mock_read_events):
    launcher = EmrPySparkStepLauncher(
        region_name="",
        staging_bucket="",
        staging_prefix="",
        wait_for_logs=False,
        action_on_failure="",
        cluster_id="",
        spark_config={},
        local_job_package_path="",
        deploy_local_job_package=False,
    )
    yielded_events = list(
        launcher.wait_for_completion(mock.MagicMock(), None, None, None, None, check_interval=0)
    )
    assert yielded_events == [event.dagster_event for event in EVENTS if event.is_dagster_event]


def test_emr_pyspark_step_launcher_legacy_arguments():
    mock_config = {
        "local_job_package_path": os.path.abspath(os.path.dirname(__file__)),
        "cluster_id": "123",
        "staging_bucket": "bucket",
        "region_name": "us-west-1",
    }

    with pytest.raises(Exception):
        emr_pyspark_step_launcher(
            build_init_resource_context(
                config={
                    **mock_config,
                    "local_pipeline_package_path": "path",
                }
            )
        )

    with pytest.raises(Exception):
        emr_pyspark_step_launcher(
            build_init_resource_context(
                config={
                    **mock_config,
                    "deploy_local_job_package": True,
                    "deploy_local_pipeline_package": True,
                }
            )
        )

    with pytest.raises(Exception):
        emr_pyspark_step_launcher(
            build_init_resource_context(
                config={
                    **mock_config,
                    "s3_job_package_path": "path",
                    "s3_pipeline_package_path": "path",
                }
            )
        )

    assert emr_pyspark_step_launcher(
        build_init_resource_context(
            config={
                **mock_config,
                "deploy_local_job_package": True,
            }
        )
    )



import os
import subprocess
import sys
from unittest import mock

import pytest
from dagster import reconstructable
from dagster._core.definitions.decorators import op
from dagster._core.definitions.decorators.graph_decorator import graph
from dagster._core.definitions.executor_definition import in_process_executor
from dagster._core.definitions.input import In
from dagster._core.definitions.no_step_launcher import no_step_launcher
from dagster._core.definitions.output import Out
from dagster._core.errors import DagsterSubprocessError
from dagster._core.execution.api import execute_job
from dagster._core.test_utils import instance_for_test
from dagster._utils.merger import deep_merge_dicts
from dagster._utils.test import create_test_pipeline_execution_context
from dagster_pyspark import DataFrame, pyspark_resource
from moto import mock_aws
from pyspark.sql import Row
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from dagster_aws.emr import EmrError, EmrJobRunner
from dagster_aws.emr.pyspark_step_launcher import EmrPySparkStepLauncher, emr_pyspark_step_launcher
from dagster_aws.s3 import s3_resource

S3_BUCKET = "dagster-scratch-80542c2"


BASE_EMR_PYSPARK_STEP_LAUNCHER_CONFIG = {
    "local_pipeline_package_path": os.path.abspath(os.path.dirname(__file__)),
    "cluster_id": os.environ.get("EMR_CLUSTER_ID"),
    "staging_bucket": S3_BUCKET,
    "region_name": "us-west-1",
    "deploy_local_pipeline_package": True,
}


@op(
    out=Out(DataFrame),
    required_resource_keys={"pyspark_step_launcher", "pyspark"},
)
def make_df_solid(context):
    schema = StructType([StructField("name", StringType()), StructField("age", IntegerType())])
    rows = [
        Row(name="John", age=19),
        Row(name="Jennifer", age=29),
        Row(name="Henry", age=50),
    ]
    return context.resources.pyspark.spark_session.createDataFrame(rows, schema)


@op(
    name="blah",
    description="this is a test",
    config_schema={"foo": str, "bar": int},
    ins={"people": In(DataFrame)},
    out=Out(DataFrame),
    required_resource_keys={"pyspark_step_launcher"},
)
def filter_df_solid(_, people):
    return people.filter(people["age"] < 30)


PROD_RESOURCES = {
    "pyspark_step_launcher": emr_pyspark_step_launcher,
    "pyspark": pyspark_resource,
    "s3": s3_resource,
}

LOCAL_RESOURCES = {
    "pyspark_step_launcher": no_step_launcher,
    "pyspark": pyspark_resource,
}


@graph
def pyspark_graph():
    filter_df_solid(make_df_solid())


def define_pyspark_job_prod():
    return pyspark_graph.to_job(resource_defs=PROD_RESOURCES)


def define_pyspark_job_local():
    return pyspark_graph.to_job(resource_defs=LOCAL_RESOURCES)


@op(
    required_resource_keys={"pyspark_step_launcher", "pyspark"},
)
def noop_op(_):
    pass


@graph
def noop_graph():
    noop_op()


def define_noop_job_prod():
    return noop_graph.to_job(resource_defs=PROD_RESOURCES, executor_def=in_process_executor)


def define_noop_job_local():
    return noop_graph.to_job(resource_defs=LOCAL_RESOURCES, executor_def=in_process_executor)


@pytest.mark.skipif(sys.version_info >= (3, 11), reason="no pyspark support on 3.11")
def test_local():
    result = define_pyspark_job_local().execute_in_process(
        run_config={"ops": {"blah": {"config": {"foo": "a string", "bar": 123}}}},
    )
    assert result.success


@mock_aws
@mock.patch("dagster_aws.emr.pyspark_step_launcher.EmrPySparkStepLauncher.read_events")
@mock.patch("dagster_aws.emr.emr.EmrJobRunner.is_emr_step_complete")
@pytest.mark.skipif(sys.version_info >= (3, 11), reason="no pyspark support on 3.11")
def test_pyspark_emr(mock_is_emr_step_complete, mock_read_events, mock_s3_bucket):
    with instance_for_test() as instance:
        with execute_job(reconstructable(define_noop_job_local), instance=instance) as result:
            mock_read_events.return_value = instance.all_logs(result.run_id)

    run_job_flow_args = dict(
        Instances={
            "InstanceCount": 1,
            "KeepJobFlowAliveWhenNoSteps": True,
            "MasterInstanceType": "c3.medium",
            "Placement": {"AvailabilityZone": "us-west-1a"},
            "SlaveInstanceType": "c3.xlarge",
        },
        JobFlowRole="EMR_EC2_DefaultRole",
        LogUri=f"s3://{mock_s3_bucket.name}/log",
        Name="cluster",
        ServiceRole="EMR_DefaultRole",
        VisibleToAllUsers=True,
    )

    # Doing cluster setup outside of a solid here, because run_job_flow is not yet plumbed through
    # to the pyspark EMR resource.
    job_runner = EmrJobRunner(region="us-west-1")
    context = create_test_pipeline_execution_context()
    cluster_id = job_runner.run_job_flow(context.log, run_job_flow_args)

    with instance_for_test() as instance:
        with execute_job(
            reconstructable(define_noop_job_prod),
            instance=instance,
            raise_on_error=True,
            run_config={
                "resources": {
                    "pyspark_step_launcher": {
                        "config": deep_merge_dicts(
                            BASE_EMR_PYSPARK_STEP_LAUNCHER_CONFIG,
                            {
                                "cluster_id": cluster_id,
                                "staging_bucket": mock_s3_bucket.name,
                            },
                        ),
                    }
                },
            },
        ) as result:
            assert result.success
            assert mock_is_emr_step_complete.called


def sync_code():
    # Sync remote dagster packages with local dagster code
    sync_code_command = [
        "rsync",
        "-av",
        "-progress",
        "--exclude='js_modules/'",
        "--exclude='.git/'",
        "--exclude='docs/'",
        "-e",
        '"ssh -i {aws_emr_pem_file}"'.format(aws_emr_pem_file=os.environ["AWS_EMR_PEM_FILE"]),
        os.environ["DAGSTER_DIR"],
        os.environ["AWS_EMR_NODE_ADDRESS"] + ":~/",
    ]
    if (
        subprocess.call(
            " ".join(sync_code_command),
            stdout=sys.stdout,
            stderr=sys.stderr,
            shell=True,
        )
        != 0
    ):
        raise DagsterSubprocessError("Failed to sync code to EMR")

    # Install dagster packages on remote node
    remote_install_dagster_packages_command = [
        "sudo",
        "python3",
        "-m",
        "pip",
        "install",
    ] + [
        token
        for package_subpath in ["dagster", "libraries/dagster-pyspark"]
        for token in ["-e", "/home/hadoop/dagster/python_modules/" + package_subpath]
    ]

    install_dagster_packages_command = [
        "ssh",
        "-i",
        os.environ["AWS_EMR_PEM_FILE"],
        os.environ["AWS_EMR_NODE_ADDRESS"],
        "'" + " ".join(remote_install_dagster_packages_command) + "'",
    ]
    if (
        subprocess.call(
            " ".join(install_dagster_packages_command),
            stdout=sys.stdout,
            stderr=sys.stderr,
            shell=True,
        )
        != 0
    ):
        raise DagsterSubprocessError("Failed to install dagster packages on EMR")


@pytest.mark.skipif(
    "AWS_EMR_TEST_DO_IT_LIVE" not in os.environ,
    reason="This test is slow and requires a live EMR cluster; run only upon explicit request",
)
def test_do_it_live_emr():
    sync_code()

    with instance_for_test() as instance:
        with execute_job(
            reconstructable(define_pyspark_job_prod),
            instance=instance,
            run_config={
                "ops": {"blah": {"config": {"foo": "a string", "bar": 123}}},
                "resources": {
                    "pyspark_step_launcher": {"config": BASE_EMR_PYSPARK_STEP_LAUNCHER_CONFIG},
                },
            },
        ) as result:
            assert result.success


@mock.patch("boto3.resource")
@mock.patch("dagster_aws.emr.pyspark_step_launcher.EmrPySparkStepLauncher.wait_for_completion")
@mock.patch("dagster_aws.emr.pyspark_step_launcher.EmrPySparkStepLauncher._log_logs_from_s3")
@mock.patch("dagster._core.events.log_step_event")
@pytest.mark.skipif(sys.version_info >= (3, 11), reason="no pyspark support on 3.11")
def test_fetch_logs_on_fail(
    _mock_log_step_event, mock_log_logs, mock_wait_for_completion, _mock_boto3_resource
):
    mock_log = mock.MagicMock()
    mock_step_context = mock.MagicMock()
    mock_step_context.log = mock_log
    mock_wait_for_completion.side_effect = EmrError()

    step_launcher = EmrPySparkStepLauncher(
        region_name="",
        staging_bucket="",
        staging_prefix="",
        spark_config=None,
        action_on_failure="",
        cluster_id="",
        local_job_package_path="",
        deploy_local_job_package=False,
        wait_for_logs=True,
    )

    with pytest.raises(EmrError):
        for _ in step_launcher.wait_for_completion_and_log(None, None, None, mock_step_context):
            pass

    assert mock_log_logs.call_count == 1



from dagster._core.remote_origin import (
    RegisteredCodeLocationOrigin,
    RemoteJobOrigin,
    RemoteRepositoryOrigin,
)

from dagster_aws.ecs.utils import get_task_definition_family, sanitize_family, sanitize_tag


def test_sanitize_family():
    assert sanitize_family("abc") == "abc"
    assert sanitize_family("abc123") == "abc123"
    assert sanitize_family("abc-123") == "abc-123"
    assert sanitize_family("abc_123") == "abc_123"
    assert sanitize_family("abc 123") == "abc123"
    assert sanitize_family("abc~123") == "abc123"


def test_sanitize_tag():
    assert sanitize_tag("abc") == "abc"
    assert sanitize_tag("abc123") == "abc123"
    assert sanitize_tag("foo.bar[filename_0]") == "foo-bar-filename_0"
    assert sanitize_tag("AaBbCc") == "AaBbCc"
    assert sanitize_tag("A" * 270) == "A" * 255


def test_get_task_definition_family():
    remote_job_origin = RemoteJobOrigin(
        repository_origin=RemoteRepositoryOrigin(
            repository_name="the_repo",
            code_location_origin=RegisteredCodeLocationOrigin(location_name="the_location"),
        ),
        job_name="the_job",
    )

    assert (
        get_task_definition_family("foo", remote_job_origin)
        == "foo_the_location_66c360f2_the_repo_b9c5532e_the_job_38cc9a96"
    )


def test_long_names():
    long_job_name = "a" * 512
    long_repo_name = "b" * 512
    long_location_name = "c" * 512

    remote_job_origin = RemoteJobOrigin(
        repository_origin=RemoteRepositoryOrigin(
            repository_name=long_repo_name,
            code_location_origin=RegisteredCodeLocationOrigin(location_name=long_location_name),
        ),
        job_name=long_job_name,
    )

    assert (
        get_task_definition_family("foo", remote_job_origin)
        == f"foo_{'c' * 55}_d9023790_{'b' * 55}_3956139d_{'a' * 55}_164557fa"
    )



import boto3
import pytest
from moto import mock_aws

from dagster_aws_tests.ecs_tests.stubbed_ecs import ThreadsafeStubbedEcs


@pytest.fixture
def region():
    return "us-east-1"


@pytest.fixture
def ecs(region):
    return ThreadsafeStubbedEcs(region_name=region)


@pytest.fixture
def ec2(region):
    with mock_aws():
        yield boto3.resource("ec2", region_name=region)


@pytest.fixture
def vpc(ec2):
    return ec2.create_vpc(CidrBlock="10.0.0.0/16")


@pytest.fixture
def subnet(vpc):
    return vpc.create_subnet(CidrBlock="10.0.0.0/24")


@pytest.fixture
def security_group(vpc):
    return vpc.create_security_group(Description="test", GroupName="test")


@pytest.fixture
def secrets_manager(region):
    with mock_aws():
        yield boto3.client("secretsmanager", region_name=region)



import boto3
import pytest
from moto import mock_aws


# Make sure unit tests never connect to real AWS
@pytest.fixture(autouse=True)
def fake_aws_credentials(monkeypatch):
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "test")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "test")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "us-east-1")


@pytest.fixture
def mock_s3_resource():
    with mock_aws():
        yield boto3.resource("s3", region_name="us-east-1")


@pytest.fixture
def mock_s3_bucket(mock_s3_resource):
    yield mock_s3_resource.create_bucket(Bucket="test-bucket")


@pytest.fixture
def mock_secretsmanager_resource():
    with mock_aws():
        yield boto3.client("secretsmanager")


@pytest.fixture
def mock_ssm_client():
    with mock_aws():
        yield boto3.client("ssm")


@pytest.fixture
def mock_rds_client():
    with mock_aws():
        yield boto3.client("rds")


@pytest.fixture
def mock_rds_data_client():
    with mock_aws():
        yield boto3.client("rds-data")



import base64
import io
import json
import os
import subprocess
import sys
import tempfile
import traceback
from typing import Any

import boto3
from dagster_pipes import PipesMappingParamsLoader, PipesS3MessageWriter, open_dagster_pipes


class LambdaFunctions:
    @staticmethod
    def trunc_logs(event, context):
        sys.stdout.write("O" * 1024 * 3)
        sys.stderr.write("E" * 1024 * 3)

    @staticmethod
    def small_logs(event, context):
        print("S" * event["size"])  # noqa: T201

    @staticmethod
    def pipes_basic(event, _lambda_context):
        with open_dagster_pipes(params_loader=PipesMappingParamsLoader(event)) as dagster_context:
            dagster_context.report_asset_materialization(metadata={"meta": "data"})

    @staticmethod
    def pipes_messy_logs(event, _lambda_context):
        with open_dagster_pipes(params_loader=PipesMappingParamsLoader(event)) as dagster_context:
            # without buffering this would get lost
            dagster_context.report_asset_materialization(metadata={"meta": "data"})
            sys.stderr.write("E" * 1024 * 4)

    @staticmethod
    def pipes_s3_messages(event, _lambda_context):
        s3_client = boto3.client(
            "s3",
            region_name="us-east-1",
            endpoint_url="http://localhost:5193",
        )
        with open_dagster_pipes(
            params_loader=PipesMappingParamsLoader(event),
            message_writer=PipesS3MessageWriter(
                client=s3_client,
                interval=0.001,
            ),
        ) as dagster_context:
            dagster_context.report_asset_materialization(metadata={"meta": "data"})

    @staticmethod
    def error(event, _lambda_context):
        raise Exception("boom")


class FakeLambdaContext:
    pass


LOG_TAIL_LIMIT = 4096


class FakeLambdaClient:
    def invoke(self, **kwargs):
        # emulate lambda constraints with a subprocess invocation
        # * json serialized "Payload" result
        # * 4k log output as base64 "LogResult"

        with tempfile.TemporaryDirectory() as tempdir:
            in_path = os.path.join(tempdir, "in.json")
            out_path = os.path.join(tempdir, "out.json")
            log_path = os.path.join(tempdir, "logs")

            with open(in_path, "w") as f:
                f.write(kwargs["Payload"])

            with open(log_path, "w") as log_file:
                result = subprocess.run(
                    [
                        sys.executable,
                        os.path.join(os.path.dirname(__file__), "fake_lambda.py"),
                        kwargs["FunctionName"],
                        in_path,
                        out_path,
                    ],
                    check=False,
                    env={},  # env vars part of lambda fn definition, can't vary at runtime
                    stdout=log_file,
                    stderr=log_file,
                )

            response: dict[str, Any] = {}

            if result.returncode == 42:
                response["FunctionError"] = "Unhandled"

            elif result.returncode != 0:
                with open(log_path) as f:
                    print(f.read())  # noqa: T201
                result.check_returncode()

            with open(out_path, "rb") as f:
                payload = io.BytesIO(f.read())

            response["Payload"] = payload

            if kwargs.get("LogType") == "Tail":
                logs_len = os.path.getsize(log_path)
                with open(log_path, "rb") as log_file:
                    if logs_len > LOG_TAIL_LIMIT:
                        log_file.seek(-LOG_TAIL_LIMIT, os.SEEK_END)

                    outro = log_file.read()

                log_result = base64.encodebytes(outro)

                response["LogResult"] = log_result

        return response


if __name__ == "__main__":
    assert len(sys.argv) == 4, "python fake_lambda.py <fn_name> <in_path> <out_path>"
    _, fn_name, in_path, out_path = sys.argv

    event = json.load(open(in_path))
    fn = getattr(LambdaFunctions, fn_name)

    val = None
    return_code = 0
    try:
        val = fn(event, FakeLambdaContext())
    except Exception as e:
        tb = traceback.TracebackException.from_exception(e)
        if sys.version_info >= (3, 13):
            name = tb.exc_type_str.split(".")[-1]
        else:
            name = tb.exc_type.__name__ if tb.exc_type is not None else None
        val = {
            "errorMessage": str(tb),
            "errorType": name,
            "stackTrace": tb.stack.format(),
            "requestId": "fake-request-id",
        }
        return_code = 42

    with open(out_path, "w") as f:
        json.dump(val, f)

    sys.exit(return_code)



from collections.abc import Iterator
from typing import TYPE_CHECKING

import boto3
import pytest
from moto.server import ThreadedMotoServer  # type: ignore  # (no type stubs)

from dagster_aws_tests.pipes_tests.utils import (
    _MOTO_SERVER_PORT,
    _MOTO_SERVER_URL,
    _S3_TEST_BUCKET,
    temp_script,
)

if TYPE_CHECKING:
    from mypy_boto3_logs import CloudWatchLogsClient


@pytest.fixture
def moto_server() -> Iterator[ThreadedMotoServer]:
    # We need to use the moto server for cross-process communication
    server = ThreadedMotoServer(port=_MOTO_SERVER_PORT)  # on localhost:5000 by default
    server.start()
    yield server
    server.stop()


@pytest.fixture
def s3_client(moto_server: ThreadedMotoServer):
    client = boto3.client("s3", region_name="us-east-1", endpoint_url=_MOTO_SERVER_URL)
    client.create_bucket(Bucket=_S3_TEST_BUCKET)
    return client


@pytest.fixture
def cloudwatch_client(moto_server, s3_client) -> "CloudWatchLogsClient":
    return boto3.client("logs", region_name="us-east-1", endpoint_url=_MOTO_SERVER_URL)


@pytest.fixture
def external_script_default_components() -> Iterator[str]:
    # This is called in an external process and so cannot access outer scope
    def script_fn():
        import os
        import time

        from dagster_pipes import open_dagster_pipes

        with open_dagster_pipes() as context:
            context.log.info("hello world")
            context.report_asset_materialization(
                metadata={"bar": {"raw_value": context.get_extra("bar"), "type": "md"}},
                data_version="alpha",
            )
            context.report_asset_check(
                "foo_check",
                passed=True,
                severity="WARN",
                metadata={
                    "meta_1": 1,
                    "meta_2": {"raw_value": "foo", "type": "text"},
                },
            )
            time.sleep(float(os.getenv("SLEEP_SECONDS", "0.1")))

    with temp_script(script_fn) as script_path:
        yield script_path


@pytest.fixture
def external_script() -> Iterator[str]:
    # This is called in an external process and so cannot access outer scope
    def script_fn():
        import os  # noqa
        import time

        import boto3
        from dagster_pipes import PipesS3ContextLoader, PipesS3MessageWriter, open_dagster_pipes

        # this url has to be hardcoded because it's an external script
        client = boto3.client("s3", region_name="us-east-1", endpoint_url="http://localhost:5193")
        context_loader = PipesS3ContextLoader(client=client)
        message_writer = PipesS3MessageWriter(client, interval=0.001)

        with open_dagster_pipes(
            context_loader=context_loader, message_writer=message_writer
        ) as context:
            context.log.info("hello world")
            time.sleep(0.1)  # sleep to make sure that we encompass multiple intervals for S3 IO
            context.report_asset_materialization(
                metadata={"bar": {"raw_value": context.get_extra("bar"), "type": "md"}},
                data_version="alpha",
            )
            context.report_asset_check(
                "foo_check",
                passed=True,
                severity="WARN",
                metadata={
                    "meta_1": 1,
                    "meta_2": {"raw_value": "foo", "type": "text"},
                },
            )

    with temp_script(script_fn) as script_path:
        yield script_path



from dagster_aws.ecs.tasks import DagsterEcsTaskDefinitionConfig


def test_create_dagster_task_definition_dict():
    task_config = DagsterEcsTaskDefinitionConfig(
        family="my_task",
        image="foo_image:bar",
        container_name="run",
        command=["dagster", "api", "execute_run"],
        execution_role_arn="fake-role",
        log_configuration={
            "logDriver": "awslogs",
            "options": {
                "awslogs-group": "my-log-group",
                "awslogs-region": "us_east_=",
                "awslogs-stream-prefix": "my_task",
            },
        },
        task_role_arn="fake-task-role",
        environment=[
            {
                "name": "FOO_ENV_VAR",
                "value": "BAR_VALUE",
            }
        ],
        secrets=[
            {
                "name": "FOO_SECRET",
                "valueFrom": "BAR_SECRET_VALUE",
            }
        ],
        sidecars=[
            {
                "name": "DatadogAgent",
                "image": "public.ecr.aws/datadog/agent:latest",
                "environment": [
                    {"name": "ECS_FARGATE", "value": "true"},
                ],
            }
        ],
        requires_compatibilities=["EC2"],
    )

    assert task_config.task_definition_dict() == {
        "family": "my_task",
        "requiresCompatibilities": ["EC2"],
        "networkMode": "awsvpc",
        "containerDefinitions": [
            {
                "name": "run",
                "image": "foo_image:bar",
                "environment": [{"name": "FOO_ENV_VAR", "value": "BAR_VALUE"}],
                "command": ["dagster", "api", "execute_run"],
                "logConfiguration": {
                    "logDriver": "awslogs",
                    "options": {
                        "awslogs-group": "my-log-group",
                        "awslogs-region": "us_east_=",
                        "awslogs-stream-prefix": "my_task",
                    },
                },
                "secrets": [{"name": "FOO_SECRET", "valueFrom": "BAR_SECRET_VALUE"}],
            },
            {
                "name": "DatadogAgent",
                "image": "public.ecr.aws/datadog/agent:latest",
                "environment": [
                    {"name": "ECS_FARGATE", "value": "true"},
                ],
            },
        ],
        "executionRoleArn": "fake-role",
        "cpu": "256",
        "memory": "512",
        "taskRoleArn": "fake-task-role",
    }

    new_cpu_mem = task_config._replace(cpu="512", memory="1024")

    assert new_cpu_mem.task_definition_dict()["cpu"] == "512"
    assert new_cpu_mem.task_definition_dict()["memory"] == "1024"



import sys
import tempfile
import time
import warnings
from dataclasses import dataclass
from subprocess import PIPE, Popen
from typing import Literal

import boto3


@dataclass
class SimulatedJobRun:
    popen: Popen
    job_run_id: str
    log_group: str
    local_script: tempfile._TemporaryFileWrapper
    stopped: bool = False


class LocalGlueMockClient:
    def __init__(
        self,
        aws_endpoint_url: str,  # usually received from moto
        s3_client: boto3.client,  # pyright: ignore (reportGeneralTypeIssues)
        glue_client: boto3.client,  # pyright: ignore (reportGeneralTypeIssues)
        pipes_messages_backend: Literal["s3", "cloudwatch"],
        cloudwatch_client: "boto3.client | None" = None,  # pyright: ignore (reportGeneralTypeIssues)
    ):
        """This class wraps moto3 clients for S3 and Glue, and provides a way to "run" Glue jobs locally.
        This is necessary because moto3 does not actually run anything when you start a Glue job, so we won't be able
        to receive any Dagster messages from it.
        If pipes_messages_backend is configured to be CloudWatch, it also uploads stderr and stdout logs to CloudWatch
        as if this has been done by Glue.

        Once the job is submitted, it is being executed in a separate process to mimic Glue behavior.
        Once the job status is requested, the process is checked for its status and the result is returned.
        """
        self.aws_endpoint_url = aws_endpoint_url
        self.s3_client = s3_client
        self.glue_client = glue_client
        self.pipes_messages_backend = pipes_messages_backend
        self.cloudwatch_client = cloudwatch_client

        self.process = None  # jobs will be executed in a separate process

        self._job_runs: dict[str, SimulatedJobRun] = {}  # mapping of JobRunId to SimulatedJobRun

    @property
    def meta(self):
        return self.glue_client.meta

    def get_job_run(self, JobName: str, RunId: str):
        # get original response
        response = self.glue_client.get_job_run(JobName=JobName, RunId=RunId)

        # check if status override is set
        simulated_job_run = self._job_runs[RunId]

        if simulated_job_run.stopped:
            response["JobRun"]["JobRunState"] = "STOPPED"
            return response

        # check if popen has completed
        if simulated_job_run.popen.poll() is not None:
            simulated_job_run.popen.wait()
            # check status code
            if simulated_job_run.popen.returncode == 0:
                response["JobRun"]["JobRunState"] = "SUCCEEDED"
            else:
                response["JobRun"]["JobRunState"] = "FAILED"
                _, stderr = simulated_job_run.popen.communicate()
                response["JobRun"]["ErrorMessage"] = stderr.decode()

            # upload logs to cloudwatch
            if self.pipes_messages_backend == "cloudwatch":
                self._upload_logs_to_cloudwatch(RunId)
        else:
            response["JobRun"]["JobRunState"] = "RUNNING"

        return response

    def start_job_run(self, JobName: str, Arguments: dict[str, str] | None, **kwargs):
        params = {
            "JobName": JobName,
        }

        if Arguments:
            params["Arguments"] = Arguments  # type: ignore

        script_s3_path = self.glue_client.get_job(JobName=JobName)["Job"]["Command"][
            "ScriptLocation"
        ]
        bucket = script_s3_path.split("/")[2]
        key = "/".join(script_s3_path.split("/")[3:])

        # mock the job run with moto
        response = self.glue_client.start_job_run(**params)
        job_run_id = response["JobRunId"]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            f = tempfile.NamedTemporaryFile(
                delete=False
            )  # we will close this file later during garbage collection
        # load the S3 script to a local file
        self.s3_client.download_file(bucket, key, f.name)

        # execute the script in a separate process
        args = []
        for key, val in (Arguments or {}).items():
            args.append(key)
            args.append(val)
        popen = Popen(
            [sys.executable, f.name, *args],
            env={
                "AWS_ENDPOINT_URL": self.aws_endpoint_url,
                "TESTING_PIPES_MESSAGES_BACKEND": self.pipes_messages_backend,
            },
            stdout=PIPE,
            stderr=PIPE,
        )

        # record execution metadata for later use
        self._job_runs[job_run_id] = SimulatedJobRun(
            popen=popen,
            job_run_id=job_run_id,
            log_group=self.glue_client.get_job_run(JobName=JobName, RunId=job_run_id)["JobRun"][
                "LogGroupName"
            ],
            local_script=f,
        )

        return response

    def batch_stop_job_run(self, JobName: str, JobRunIds: list[str]):
        for job_run_id in JobRunIds:
            if simulated_job_run := self._job_runs.get(job_run_id):
                simulated_job_run.popen.terminate()
                simulated_job_run.stopped = True
                self._upload_logs_to_cloudwatch(job_run_id)

    def _upload_logs_to_cloudwatch(self, job_run_id: str):
        log_group = self._job_runs[job_run_id].log_group
        stdout, stderr = self._job_runs[job_run_id].popen.communicate()

        if self.pipes_messages_backend == "cloudwatch":
            assert self.cloudwatch_client is not None, (
                "cloudwatch_client has to be provided with cloudwatch messages backend"
            )

            assert self.cloudwatch_client is not None, (
                "cloudwatch_client has to be provided with cloudwatch messages backend"
            )

            try:
                self.cloudwatch_client.create_log_group(
                    logGroupName=f"{log_group}/output",
                )
            except self.cloudwatch_client.exceptions.ResourceAlreadyExistsException:
                pass

            try:
                self.cloudwatch_client.create_log_stream(
                    logGroupName=f"{log_group}/output",
                    logStreamName=job_run_id,
                )
            except self.cloudwatch_client.exceptions.ResourceAlreadyExistsException:
                pass

            for out in [stderr, stdout]:  # Glue routes both stderr and stdout to /output
                for line in out.decode().split(
                    "\n"
                ):  # uploading log lines one by one is good enough for tests
                    if line:
                        self.cloudwatch_client.put_log_events(
                            logGroupName=f"{log_group}/output",
                            logStreamName=job_run_id,
                            logEvents=[
                                {"timestamp": int(time.time() * 1000), "message": str(line)}
                            ],
                        )
        time.sleep(
            0.01
        )  # make sure the logs will be properly filtered by ms timestamp when accessed next time

    def __del__(self):
        # cleanup local script paths
        for job_run in self._job_runs.values():
            job_run.local_script.close()



import os
import uuid
from unittest import mock

import boto3
import psycopg2
import pytest
from dagster._core.definitions.decorators import op
from dagster._utils.test import wrap_op_in_graph_and_execute

from dagster_aws.redshift import (
    FakeRedshiftClient,
    FakeRedshiftClientResource,
    RedshiftClientResource,
    fake_redshift_resource,
    redshift_resource,
)

REDSHIFT_ENV = {
    "resources": {
        "redshift": {
            "config": {
                "host": "foo",
                "port": 5439,
                "user": "dagster",
                "password": "baz",
                "database": "dev",
            }
        }
    }
}

QUERY_RESULT = [(1,)]


def mock_execute_query_conn(*_args, **_kwargs):
    cursor_mock = mock.MagicMock(rowcount=1)
    cursor_mock.fetchall.return_value = QUERY_RESULT
    conn = mock.MagicMock(is_conn="yup")
    conn.cursor.return_value.__enter__.return_value = cursor_mock
    m = mock.MagicMock()
    m.return_value.__enter__.return_value = conn
    m.return_value = conn
    return m


@op(required_resource_keys={"redshift"})
def single_redshift_solid(context):
    assert context.resources.redshift
    return context.resources.redshift.execute_query("SELECT 1", fetch_results=True)


@op
def single_redshift_solid_pythonic(redshift: RedshiftClientResource):
    return redshift.get_client().execute_query("SELECT 1", fetch_results=True)


@op(required_resource_keys={"redshift"})
def multi_redshift_solid(context):
    assert context.resources.redshift
    return context.resources.redshift.execute_queries(
        ["SELECT 1", "SELECT 1", "SELECT 1"], fetch_results=True
    )


@op
def multi_redshift_solid_pythonic(redshift: RedshiftClientResource):
    return redshift.get_client().execute_queries(
        ["SELECT 1", "SELECT 1", "SELECT 1"], fetch_results=True
    )


@mock.patch("psycopg2.connect", new_callable=mock_execute_query_conn)
def test_single_select(redshift_connect):
    result = wrap_op_in_graph_and_execute(
        single_redshift_solid,
        run_config=REDSHIFT_ENV,
        resources={"redshift": redshift_resource},
    )
    redshift_connect.assert_called_once_with(
        host="foo",
        port=5439,
        user="dagster",
        password="baz",
        database="dev",
        connect_timeout=5,
        sslmode="require",
    )

    assert result.success
    assert result.output_value() == QUERY_RESULT


@mock.patch("psycopg2.connect", new_callable=mock_execute_query_conn)
def test_single_select_pythonic_resource(redshift_connect) -> None:
    result = wrap_op_in_graph_and_execute(
        single_redshift_solid_pythonic,
        run_config=REDSHIFT_ENV,
        resources={"redshift": RedshiftClientResource.configure_at_launch()},
    )
    redshift_connect.assert_called_once_with(
        host="foo",
        port=5439,
        user="dagster",
        password="baz",
        database="dev",
        connect_timeout=5,
        sslmode="require",
    )

    assert result.success
    assert result.output_value() == QUERY_RESULT


@mock.patch("psycopg2.connect", new_callable=mock_execute_query_conn)
def test_multi_select(_redshift_connect):
    result = wrap_op_in_graph_and_execute(
        multi_redshift_solid,
        run_config=REDSHIFT_ENV,
        resources={"redshift": redshift_resource},
    )
    assert result.success
    assert result.output_value() == [QUERY_RESULT] * 3


@mock.patch("psycopg2.connect", new_callable=mock_execute_query_conn)
def test_multi_select_pythonic(_redshift_connect) -> None:
    result = wrap_op_in_graph_and_execute(
        multi_redshift_solid_pythonic,
        run_config=REDSHIFT_ENV,
        resources={"redshift": RedshiftClientResource.configure_at_launch()},
    )
    assert result.success
    assert result.output_value() == [QUERY_RESULT] * 3


def test_fake_redshift() -> None:
    fake_resources = {"redshift": fake_redshift_resource}

    result = wrap_op_in_graph_and_execute(
        single_redshift_solid, run_config=REDSHIFT_ENV, resources=fake_resources
    )
    assert result.success
    assert result.output_value() == FakeRedshiftClient.QUERY_RESULT

    result = wrap_op_in_graph_and_execute(
        multi_redshift_solid, run_config=REDSHIFT_ENV, resources=fake_resources
    )
    assert result.success
    assert result.output_value() == [FakeRedshiftClient.QUERY_RESULT] * 3


def test_fake_redshift_pythonic() -> None:
    fake_resources = {"redshift": FakeRedshiftClientResource.configure_at_launch()}

    result = wrap_op_in_graph_and_execute(
        single_redshift_solid_pythonic, run_config=REDSHIFT_ENV, resources=fake_resources
    )
    assert result.success
    assert result.output_value() == FakeRedshiftClient.QUERY_RESULT

    result = wrap_op_in_graph_and_execute(
        multi_redshift_solid_pythonic, run_config=REDSHIFT_ENV, resources=fake_resources
    )
    assert result.success
    assert result.output_value() == [FakeRedshiftClient.QUERY_RESULT] * 3


REDSHIFT_CREATE_TABLE_QUERY = """CREATE TABLE IF NOT EXISTS VENUE1(
VENUEID SMALLINT,
VENUENAME VARCHAR(100),
VENUECITY VARCHAR(30),
VENUESTATE CHAR(2),
VENUESEATS INTEGER
) DISTSTYLE EVEN;
"""

REDSHIFT_LOAD_FILE_CONTENTS = b"""
7|BMO Field|Toronto|ON|0
16|TD Garden|Boston|MA|0
23|The Palace of Auburn Hills|Auburn Hills|MI|0
28|American Airlines Arena|Miami|FL|0
37|Staples Center|Los Angeles|CA|0
42|FedExForum|Memphis|TN|0
52|PNC Arena|Raleigh|NC  ,25   |0
59|Scotiabank Saddledome|Calgary|AB|0
66|SAP Center|San Jose|CA|0
73|Heinz Field|Pittsburgh|PA|65050
""".strip()

REDSHIFT_FAILED_LOAD_QUERY = """
SELECT le.query,
TRIM(le.err_reason) AS err_reason,
TRIM(le.filename) AS filename,
le.line_number AS line_number,
le.raw_line AS raw_line,
le.raw_field_value AS raw_value
FROM stl_load_errors le
WHERE le.query
AND le.query = pg_last_copy_id()
LIMIT 1;
"""


@pytest.mark.skipif(
    "AWS_REDSHIFT_TEST_DO_IT_LIVE" not in os.environ,
    reason="This test only works with a live Redshift cluster",
)
def test_live_redshift(s3_bucket):
    """Test live redshift instance.

    This test is based on:

    https://aws.amazon.com/premiumsupport/knowledge-center/redshift-stl-load-errors/

    Requires the following environment variables:

    AWS_ACCOUNT_ID - AWS account ID to use
    REDSHIFT_LOAD_IAM_ROLE - IAM role to use for Redshift load
    REDSHIFT_ENDPOINT - Redshift URL
    REDSHIFT_PASSWORD - Redshift password

    """
    # Put file to load on S3
    file_key = uuid.uuid4().hex
    client = boto3.client("s3")
    client.put_object(Body=REDSHIFT_LOAD_FILE_CONTENTS, Bucket=s3_bucket, Key=file_key)

    @op(required_resource_keys={"redshift"})
    def query(context):
        assert context.resources.redshift

        # First, create table:
        context.resources.redshift.execute_query(REDSHIFT_CREATE_TABLE_QUERY)

        def error_callback(error, cursor, _log):
            assert (
                str(error).strip()
                == "Load into table 'venue1' failed.  Check 'stl_load_errors' system table for"
                " details."
            )
            cursor.execute(REDSHIFT_FAILED_LOAD_QUERY)
            res = cursor.fetchall()
            assert res[0][1] == "Char length exceeds DDL length"
            assert res[0][2] == f"s3://{s3_bucket}/{file_key}"
            assert res[0][3] == 7
            assert res[0][4].strip() == "52|PNC Arena|Raleigh|NC  ,25   |0"
            assert res[0][5].strip() == "NC  ,25"
            raise error

        return context.resources.redshift.execute_query(
            """COPY venue1 FROM 's3://{s3_bucket}/{file_key}'
            IAM_ROLE 'arn:aws:iam::{AWS_ACCOUNT_ID}:role/{REDSHIFT_LOAD_IAM_ROLE}'
            DELIMITER '|';
            """.format(
                s3_bucket=s3_bucket,
                file_key=file_key,
                AWS_ACCOUNT_ID=os.environ["AWS_ACCOUNT_ID"],
                REDSHIFT_LOAD_IAM_ROLE=os.environ["REDSHIFT_LOAD_IAM_ROLE"],
            ),
            fetch_results=True,
            error_callback=error_callback,
        )

    with pytest.raises(psycopg2.InternalError):
        wrap_op_in_graph_and_execute(
            query,
            run_config={
                "resources": {
                    "redshift": {
                        "config": {
                            "host": {"env": "REDSHIFT_ENDPOINT"},
                            "port": 5439,
                            "user": "dagster",
                            "password": {"env": "REDSHIFT_PASSWORD"},
                            "database": "dev",
                        }
                    }
                }
            },
            resources={"redshift": redshift_resource},
        )
